"""Define reusable pytest fixtures."""
import tempfile
from subprocess import run
from types import SimpleNamespace

import pytest

from yamlpath.wrappers import ConsolePrinter
from yamlpath.eyaml import EYAMLProcessor

# Implied constants
EYAML_PRIVATE_KEY_FILENAME = "private_key.pkcs7.pem"
EYAML_PUBLIC_KEY_FILENAME = "public_key.pkcs7.pem"

# pylint: disable=locally-disabled,invalid-name
requireseyaml = pytest.mark.skipif(
    EYAMLProcessor.get_eyaml_executable("eyaml") is None
    , reason="The 'eyaml' command must be installed and accessible on the PATH"
        + " to test and use EYAML features.  Try:  'gem install hiera-eyaml'"
        + " after intalling ruby and rubygems."
)

@pytest.fixture
def quiet_logger():
    """Returns a quiet ConsolePrinter."""
    args = SimpleNamespace(verbose=False, quiet=True, debug=False)
    return ConsolePrinter(args)

@pytest.fixture
def info_warn_logger():
    """Returns a quiet ConsolePrinter."""
    args = SimpleNamespace(verbose=False, quiet=False, debug=False)
    return ConsolePrinter(args)

@pytest.fixture(scope="session")
def old_eyaml_keys(tmp_path_factory):
    """Creates temporary keys for encryption/decryption tests."""
    old_key_path_name = "old-keys"
    old_key_dir = tmp_path_factory.mktemp(old_key_path_name)
    old_private_key_file = old_key_dir / EYAML_PRIVATE_KEY_FILENAME
    old_public_key_file = old_key_dir / EYAML_PUBLIC_KEY_FILENAME

    old_private_key = r"""-----BEGIN RSA PRIVATE KEY-----
MIIEpAIBAAKCAQEA1BuytnsdHdt6NkNfLoGJIlf9hrWux8raPP3W57cONh2MrQ6d
aoAX+L+igTSjvSTI6oxsO0dqdYXZO1+rOK3gI9OnZQhkCjq9IRoWx7AIvM7skaD0
Lne9YsvA7mGY/z9lm3IALI6OBVV5k6xnBR2PVi6A7FnDm0CRLit2Bn9eHLN3k4oL
S/ynxgXBmWWgnKtJNJwGmeD5PwzJfXCcJ3kPItiktFizJZoPmAlBP7LIzamlfSXV
VoniRs45aGrTGpmZSdvossL41KBCYJGjP+lIL/UpDJHBeiuqVQoDl4/UZqb5xF9S
C2p2Of21fmZmj4uUAT5FPtKMKCspmLWQeUEfiwIDAQABAoIBAEyXR9wu7p+mbiYE
A+22Jr+5CDpJhrhsXovhmWWIq2ANIYwoF92qLX3MLTD8whd9nfNcC4UIT7/qOjv/
WsOXvbUSK4MHGaC7/ylh01H+Fdmf2rrnZOUWpdN0AdHSej3JNbaA3uE4BL6WU9Vo
TrcBKo4TMsilzUVVdlc2qGLGQUSZPLnIJWMLQIdCe2kZ9YvUlGloD4uGT3RH6+vH
TOtXqDgLGS/79K0rnflnBsUBkXpukxzOcTRHxR0s7SJ2XCB0JfdLWfR6X1nzM4mh
rn/m2nzEOG9ICe5hoHqAEZ/ezKd/jwxMg1YMZnGAzDMw7/UPWo87wgVdxxOHOsHG
v/pK+okCgYEA/SToT82qtvWIiUTbkdDislGhTet2eEu2Bi9wCxAUQn045Y812r9d
TvJyfKJyvvpxzejaecJb8oOXyypMDay7aPOVBI1E2OqfImxF8pJ0QqejAUCinXrj
KnV72L/hjTavivWq1vHZYXSxufAMG0C7UeztwkOfk85N3wuuYYWYrc0CgYEA1oBG
2fQ0PXDyionE3c4bpRGZMJxD+3RFRMCJiE+xheRR22ObSDLUH123ZGmU0m2FTS9O
M+GJbZgdV1Oe0EJ5rWfzFYUmVJIreH+oQWaY/HMkoe705LAMcPyd0bSjRVWWiz+l
anIGjj5HaPSI7XFqdQu7j3ww67k4BBAca8++arcCgYA/cIhnt3sY7t+QxxjfqiGl
3p82D9RYwWCUnD7QBu+M2iTwIru0XlDcABaA9ZUcF1d96uUVroesdx4LZEY7BxbQ
bnrh8SVX1zSaQ9gjumA4dBp9rd0S6kET2u12nF/CK/pCMN7njySTL9N6bZYbHlXT
ajULgjbzq7gINb01420n4QKBgQCqu0epy9qY3QHwi2ALPDZ82NkZ/AeQaieIZcgS
m3wtmmIdQdcjTHHS1YFXh0JRi6MCoJiaavY8KUuRapmKIp8/CvJNOsIbpoy7SMDf
7Y3vwqZxzgVW0VnVxPzJIgKi+VDuXSaI52GYbrHgNGOYuyGFMGWF+8/kkHSppzk4
Bw8FWQKBgQCo/7cV19n3e7ZlZ/aGhIOtSCTopMBV8/9PIw+s+ZDiQ7vRSj9DwkAQ
+x97V0idgh16tnly0xKvGCGTQR7qDsDTjHmmF4LZUGjcq7pHsTi/umCM/toE+BCk
7ayr+G0DWr5FjhQ7uCt2Rz1NKcj6EkDcM1WZxkDLAvBXjlj+T+eqtQ==
-----END RSA PRIVATE KEY-----
"""
    old_public_key = r"""-----BEGIN CERTIFICATE-----
MIIC2TCCAcGgAwIBAgIBATANBgkqhkiG9w0BAQsFADAAMCAXDTE5MDUwNzE4MDAw
NVoYDzIwNjkwNDI0MTgwMDA1WjAAMIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIB
CgKCAQEA1BuytnsdHdt6NkNfLoGJIlf9hrWux8raPP3W57cONh2MrQ6daoAX+L+i
gTSjvSTI6oxsO0dqdYXZO1+rOK3gI9OnZQhkCjq9IRoWx7AIvM7skaD0Lne9YsvA
7mGY/z9lm3IALI6OBVV5k6xnBR2PVi6A7FnDm0CRLit2Bn9eHLN3k4oLS/ynxgXB
mWWgnKtJNJwGmeD5PwzJfXCcJ3kPItiktFizJZoPmAlBP7LIzamlfSXVVoniRs45
aGrTGpmZSdvossL41KBCYJGjP+lIL/UpDJHBeiuqVQoDl4/UZqb5xF9SC2p2Of21
fmZmj4uUAT5FPtKMKCspmLWQeUEfiwIDAQABo1wwWjAPBgNVHRMBAf8EBTADAQH/
MB0GA1UdDgQWBBTUHb3HX8dBfYFL1J1sCv+uCum5AzAoBgNVHSMEITAfgBTUHb3H
X8dBfYFL1J1sCv+uCum5A6EEpAIwAIIBATANBgkqhkiG9w0BAQsFAAOCAQEAcw+0
dfHSLNAZD95G2pDnT2qShjmdLdbrDQhAXWhLeoWpXsKvC0iUyQaOF9ckl++tHM2g
ejm1vEOrZ+1uXK3qnMXPF99Wet686OhyoDt262Mt3wzGHNijAHEvQtjap8ZIwfOM
zFTvjmOlUScqF/Yg+htcGnJdQhWIrsD+exiY5Kz2IMtuW+yWLLP8bY5vPg6qfrp2
4VVJ3Md1gdSownd1Au5tqPXm6VfSgLiCm9iDPVsjDII9h8ydate1d2TBHPup+4tN
JZ5/muctimydC+S2oCn7ucsilxZD89N7rJjKXNfoUOGHjOEVQMa8RtZLzH2sUEaS
FktE6rH8a+8SwO+TGw==
-----END CERTIFICATE-----
"""

    with open(old_private_key_file, 'w') as key_file:
        key_file.write(old_private_key)

    with open(old_public_key_file, 'w') as key_file:
        key_file.write(old_public_key)

    return (old_private_key_file, old_public_key_file)

@requireseyaml
@pytest.fixture(scope="session")
def new_eyaml_keys(tmp_path_factory):
    """Creates temporary keys for encryption/decryption tests."""

    new_key_path_name = "new-keys"
    new_key_dir = tmp_path_factory.mktemp(new_key_path_name)
    new_private_key_file = new_key_dir / EYAML_PRIVATE_KEY_FILENAME
    new_public_key_file = new_key_dir / EYAML_PUBLIC_KEY_FILENAME

    run(
        "{} createkeys --pkcs7-private-key={} --pkcs7-public-key={}"
        .format(
            EYAMLProcessor.get_eyaml_executable("eyaml"),
            new_private_key_file,
            new_public_key_file
        )
        .split(),
        check=True
    )

    return (new_private_key_file, new_public_key_file)

def create_temp_yaml_file(tmp_path_factory, content):
    """Creates a test YAML input file."""
    fhnd = tempfile.NamedTemporaryFile(mode='w',
                                       dir=tmp_path_factory.getbasetemp(),
                                       suffix='.yaml',
                                       delete=False)
    fhnd.write(content)
    return fhnd.name

@pytest.fixture(scope="session")
def imparsible_yaml_file(tmp_path_factory):
    """
    Creates a YAML file that causes a ParserError when read by ruamel.yaml.
    """
    content = '''{"json": "is YAML", "but_bad_json": "isn't anything!"'''
    return create_temp_yaml_file(tmp_path_factory, content)

@pytest.fixture(scope="session")
def badsyntax_yaml_file(tmp_path_factory):
    """
    Creates a YAML file that causes a ScannerError when read by ruamel.yaml.
    """
    content = """---
    # This YAML content contains a critical syntax error
    & bad_anchor: is bad
    """
    return create_temp_yaml_file(tmp_path_factory, content)

@pytest.fixture(scope="session")
def badcmp_yaml_file(tmp_path_factory):
    """
    Creates a YAML file that causes a ComposerError when read by ruamel.yaml.
    """
    content = """---
    # This YAML file is improperly composed
    this is a parsing error: *no such capability
    """
    return create_temp_yaml_file(tmp_path_factory, content)



"""
Implements the ArrayDiffOpts enumeration.

Copyright 2020 William W. Kimball, Jr. MBA MSIS
"""
from enum import Enum, auto
from typing import List


class ArrayDiffOpts(Enum):
    """
    Supported Array (AKA: List) Diff Options.

    Options include:

    `POSITION`
        Array elements are compared based on their ordinal position in each
        document.

    `VALUE`
        Array alements are synchronized by value before being compared.
    """

    POSITION = auto()
    VALUE = auto()

    def __str__(self) -> str:
        """
        Stringify one instance of this enumeration.

        Parameters:  N/A

        Returns:  (str) String value of this enumeration.

        Raises:  N/A
        """
        return str(self.name).lower()

    @staticmethod
    def get_names() -> List[str]:
        """
        Get all upper-cased entry names for this enumeration.

        Parameters:  N/A

        Returns:  (List[str]) Upper-case names from this enumeration

        Raises:  N/A
        """
        return [entry.name.upper() for entry in ArrayDiffOpts]

    @staticmethod
    def get_choices() -> List[str]:
        """
        Get all entry names with symbolic representations for this enumeration.

        All returned entries are lower-cased.

        Parameters:  N/A

        Returns:  (List[str]) Lower-case names and symbols from this
            enumeration

        Raises:  N/A
        """
        names = [l.lower() for l in ArrayDiffOpts.get_names()]
        choices = list(set(names))
        choices.sort()
        return choices

    @staticmethod
    def from_str(name: str) -> "ArrayDiffOpts":
        """
        Convert a string value to a value of this enumeration, if valid.

        Parameters:
            1. name (str) The name to convert

        Returns:  (ArrayDiffOpts) the converted enumeration value

        Raises:
            - `NameError` when name doesn't match any enumeration values
        """
        check: str = str(name).upper()
        if check in ArrayDiffOpts.get_names():
            return ArrayDiffOpts[check]
        raise NameError(
            "ArrayDiffOpts has no such item:  {}"
            .format(name))



"""
Defines a custom type for YAML Path segments.

Copyright 2020 William W. Kimball, Jr. MBA MSIS
"""
from typing import Tuple

from yamlpath.enums import PathSegmentTypes
from yamlpath.types import PathAttributes

PathSegment = Tuple[PathSegmentTypes, PathAttributes]



"""
Enable users to change YAML/Compatible files using YAML Paths.

Changes one or more values in a YAML file at a specified YAML Path.  Matched
values can be checked before they are replaced to mitigate accidental change.
When matching singular results, the value can be archived to another key
before it is replaced.  Further, EYAML can be employed to encrypt the new
values and/or decrypt old values before checking them.

Copyright 2018, 2019, 2020 William W. Kimball, Jr. MBA MSIS
"""
import sys
import tempfile
import argparse
import secrets
import string
import json
from os import remove, access, R_OK
from os.path import isfile, exists
from shutil import copy2, copyfileobj
from pathlib import Path

from yamlpath import __version__ as YAMLPATH_VERSION
from yamlpath.common import Nodes, Parsers
from yamlpath import YAMLPath
from yamlpath.exceptions import YAMLPathException
from yamlpath.enums import YAMLValueFormats, PathSeperators
from yamlpath.eyaml.exceptions import EYAMLCommandException
from yamlpath.eyaml.enums import EYAMLOutputFormats
from yamlpath.eyaml import EYAMLProcessor
from yamlpath.wrappers import ConsolePrinter

def processcli():
    """Process command-line arguments."""
    parser = argparse.ArgumentParser(
        description=(
            "Changes one or more Scalar values in a YAML/JSON/Compatible"
            " document at a specified YAML Path.  Matched values can be"
            " checked before they are replaced to mitigate accidental change."
            "  When matching singular results, the value can be archived to"
            " another key before it is replaced.  Further, EYAML can be"
            " employed to encrypt the new values and/or decrypt an old value"
            " before checking it."),
        epilog=(
            "When no changes are made, no backup is created, even when"
            " -b/--backup is specified.  For more information about YAML"
            " Paths, please visit https://github.com/wwkimball/yamlpath/wiki."
            "  To report issues with this tool or to request enhancements,"
            " please visit https://github.com/wwkimball/yamlpath/issues.")
    )
    parser.add_argument("-V", "--version", action="version",
                        version="%(prog)s " + YAMLPATH_VERSION)

    required_group = parser.add_argument_group("required settings")
    required_group.add_argument(
        "-g", "--change",
        required=True,
        metavar="YAML_PATH",
        help="YAML Path where the target value is found")

    inputex_group = parser.add_argument_group("input options")
    input_group = inputex_group.add_mutually_exclusive_group()
    input_group.add_argument(
        "-a", "--value",
        help="set the new value from the command-line instead of STDIN")
    input_group.add_argument(
        "-A", "--aliasof",
        metavar="ANCHOR",
        help="set the value as a YAML Alias of an existing scalar value Anchor"
             " by its source ANCHOR YAML Path (merely copies the target value"
             " for non-YAML files)")
    input_group.add_argument(
        "-K", "--mergekey",
        metavar="ANCHOR",
        help="assign a YAML Merge Key to target Hash(es) by the source Hash's"
             " ANCHOR YAML Path (merely copies unique key-value pairs from the"
             " source Hash into the target Hash for non-YAML files)")
    input_group.add_argument(
        "-f", "--file",
        help="read the new value from file (discarding any trailing\
              new-lines)")
    input_group.add_argument(
        "-i", "--stdin", action="store_true",
        help="accept the new value from STDIN (best for sensitive data)")
    input_group.add_argument(
        "-R", "--random",
        type=int,
        metavar="LENGTH",
        help="randomly generate a replacement value of a set length")
    input_group.add_argument(
        "-N", "--null",
        action="store_true",
        help="sets the value to null")
    input_group.add_argument(
        "-D", "--delete",
        action="store_true",
        help="delete rather than change target node(s); implies"
             " --mustexist|-m")

    parser.add_argument(
        "-F", "--format",
        default="default",
        choices=[l.lower() for l in YAMLValueFormats.get_names()],
        type=str.lower,
        help="override automatic formatting of the new value")
    parser.add_argument(
        "-c", "--check",
        help="check the value before replacing it")
    parser.add_argument(
        "-s", "--saveto", metavar="YAML_PATH",
        help="save the old value to YAML_PATH before replacing it; implies\
              --mustexist")
    parser.add_argument(
        "-m", "--mustexist", action="store_true",
        help="require that the --change YAML_PATH already exist in YAML_FILE")
    parser.add_argument(
        "-b", "--backup", action="store_true",
        help="save a backup YAML_FILE with an extra .bak file-extension")
    parser.add_argument(
        "-t", "--pathsep",
        default="dot",
        choices=PathSeperators,
        metavar=PathSeperators.get_choices(),
        type=PathSeperators.from_str,
        help="indicate which YAML Path seperator to use when rendering\
              results; default=dot")

    parser.add_argument(
        "-M", "--random-from",
        metavar="CHARS",
        default=(string.ascii_uppercase +
            string.ascii_lowercase +
            string.digits),
        help="characters from which to build a value for --random; default="
             "all upper- and lower-case letters and all digits"
    )
    parser.add_argument(
        "-H", "--anchor",
        metavar="ANCHOR",
        help="when --aliasof|-A points to a value which is not already"
             " Anchored, a new Anchor with this name is created; renames an"
             " existing Anchor if already set")
    parser.add_argument(
        "-T", "--tag",
        metavar="TAG",
        help="assign a custom YAML (data-type) tag to the changed nodes; can"
             " be used without other input options to assign or change a tag")

    eyaml_group = parser.add_argument_group(
        "EYAML options", "Left unset, the EYAML keys will default to your\
         system or user defaults.  You do not need to supply a private key\
         unless you enable --check and the old value is encrypted.")
    eyaml_group.add_argument(
        "-e", "--eyamlcrypt", action="store_true",
        help="encrypt the new value using EYAML")
    eyaml_group.add_argument(
        "-x", "--eyaml", default="eyaml",
        help="the eyaml binary to use when it isn't on the PATH")
    eyaml_group.add_argument("-r", "--privatekey", help="EYAML private key")
    eyaml_group.add_argument("-u", "--publickey", help="EYAML public key")

    parser.add_argument(
        "-S", "--nostdin", action="store_true",
        help=(
            "Do not implicitly read from STDIN, even when there is no"
            " YAML_FILE with a non-TTY session"))

    noise_group = parser.add_mutually_exclusive_group()
    noise_group.add_argument(
        "-d", "--debug", action="store_true",
        help="output debugging details")
    noise_group.add_argument(
        "-v", "--verbose", action="store_true",
        help="increase output verbosity")
    noise_group.add_argument(
        "-q", "--quiet", action="store_true",
        help="suppress all output except errors")

    parser.add_argument(
        "yaml_file", metavar="YAML_FILE", nargs="?",
        help="the YAML file to update; omit or use - to read from STDIN")
    return parser.parse_args()

# pylint: disable=too-many-branches
def validateargs(args, log):
    """Validate command-line arguments."""
    has_errors = False
    in_file = args.yaml_file if args.yaml_file else ""
    in_stream_mode = in_file.strip() == "-" or (
        not in_file and not args.nostdin and not sys.stdin.isatty()
    )

    # When there is no YAML_FILE and no STDIN, there is nothing to read
    if not (in_file or in_stream_mode):
        has_errors = True
        log.error("There must be a YAML_FILE or STDIN document.")

    # One of the input group options must be specified.
    if not (
            args.value
            or args.value == ""
            or args.aliasof
            or args.mergekey
            or args.file
            or args.stdin
            or args.random
            or args.null
            or args.delete
            or args.anchor
            or args.tag
    ):
        has_errors = True
        log.error(
            "Exactly one of the following must be set:  --value, --aliasof,"
            " --mergekey, --file, --stdin, --random, --null, --delete,"
            " --anchor, or --tag")

    # --stdin cannot be used with -, explicit or implied
    if args.stdin and in_stream_mode:
        has_errors = True
        log.error(
            "Impossible to read both document and replacement value from"
            " STDIN!")

    # Remove potentially-accidental characters from any Anchor
    if args.anchor:
        args.anchor = (
            args.anchor
            .replace(" ", "")
            .replace("&", "")
            .replace("*", "")
        )

    # --anchor can be used only when --aliasof or --mergekey are set
    if args.anchor and not (args.aliasof or args.mergekey):
        has_errors = True
        log.error(
            "The --anchor|-H option may be used only when --aliasof|-A or"
            " --mergekey|-K are also set.")

    # --backup has no meaning when reading the YAML file from STDIN
    if args.backup and in_stream_mode:
        has_errors = True
        log.error(
            "The --backup|-b option applies only when reading from a file, not"
            " STDIN.")

    # When set, --saveto cannot be identical to --change
    if args.saveto and args.saveto == args.change:
        has_errors = True
        log.error(
            "Impossible to save the old value to the same YAML Path as the new"
            + " value!")

    # When set, --privatekey must be a readable file
    if args.privatekey and not (
            isfile(args.privatekey)
            and access(args.privatekey, R_OK)
    ):
        has_errors = True
        log.error(
            "EYAML private key is not a readable file:  " + args.privatekey)

    # When set, --publickey must be a readable file
    if args.publickey and not (
            isfile(args.publickey)
            and access(args.publickey, R_OK)
    ):
        has_errors = True
        log.error(
            "EYAML public key is not a readable file:  " + args.publickey)

    # When set, --random-from must have at least two characters
    if len(args.random_from) < 2:
        has_errors = True
        log.error("The pool of random CHARS must have at least 2 characters.")

    # Any set tag must have a prefix of at least one !
    if args.tag and not args.tag[0] == "!":
        args.tag = "!{}".format(args.tag)

    # When using --delete, --mustexist must also be set
    if args.delete:
        args.mustexist = True

    # When dumping the document to STDOUT, mute all non-errors
    force_verbose = args.verbose
    force_debug = args.debug
    if in_stream_mode and not (force_verbose or force_debug):
        args.quiet = True
        args.verbose = False
        args.debug = False

    if has_errors:
        sys.exit(1)

def save_to_json_file(args, log, yaml_data):
    """Save to a JSON file."""
    log.verbose("Writing changed data as JSON to {}.".format(args.yaml_file))
    with open(args.yaml_file, 'w', encoding='utf-8') as out_fhnd:
        json.dump(Parsers.jsonify_yaml_data(yaml_data), out_fhnd)

def save_to_yaml_file(args, log, yaml_parser, yaml_data, backup_file):
    """Save to a YAML file."""
    log.verbose("Writing changed data as YAML to {}.".format(args.yaml_file))
    with tempfile.TemporaryFile() as tmphnd:
        with open(args.yaml_file, 'rb') as inhnd:
            copyfileobj(inhnd, tmphnd)

        with open(args.yaml_file, 'w', encoding='utf-8') as yaml_dump:
            try:
                yaml_parser.dump(yaml_data, yaml_dump)
            # Tell pycov to ignore this block because it is impossible to
            # trigger it for ruamel.yaml versions >0.17.4 yet this project must
            # continue to support older versions of ruamel.yaml as long as OS
            # package builders continue to be dependent on them.
            except AssertionError as ex:    # pragma: no cover
                yaml_dump.close()
                tmphnd.seek(0)
                with open(args.yaml_file, 'wb') as outhnd:
                    copyfileobj(tmphnd, outhnd)

                # No sense in preserving a backup file with no changes
                if args.backup:
                    remove(backup_file)

                log.debug(
                    "yaml_set::save_to_yaml_file:  Assertion error: {}"
                    .format(ex))
                log.critical((
                    "Indeterminate assertion error encountered while"
                    + " attempting to write updated data to {}.  The original"
                    + " file content was restored.").format(args.yaml_file), 3)

def docroot_is_flow(yaml_data):
    """Determine whether a document root is in flow (JSON) style."""
    is_flow = False
    if hasattr(yaml_data, "fa"):
        is_flow = yaml_data.fa.flow_style()
    return is_flow

def write_document_as_yaml(output_file_name, yaml_data):
    """Determine whether to write out YAML (or JSON)."""
    write_yaml = True
    if docroot_is_flow(yaml_data):
        write_yaml = False

    # Allow a JSON file extension to override the inference
    if write_yaml:
        write_yaml = Path(output_file_name).suffix.lower() != ".json"

    return write_yaml

def save_to_file(args, log, yaml_parser, yaml_data, backup_file):
    """Save as YAML or JSON."""
    if write_document_as_yaml(args.yaml_file, yaml_data):
        save_to_yaml_file(args, log, yaml_parser, yaml_data, backup_file)
    else:
        save_to_json_file(args, log, yaml_data)

def write_output_document(args, log, yaml, yaml_data):
    """Write the updated document to file or STDOUT."""
    # Save a backup of the original file, if requested
    backup_file = args.yaml_file + ".bak"
    if args.backup:
        log.verbose(
            "Saving a backup of {} to {}."
            .format(args.yaml_file, backup_file))
        if exists(backup_file):
            remove(backup_file)
        copy2(args.yaml_file, backup_file)

    # Save the changed file
    if args.yaml_file.strip() == "-":
        if write_document_as_yaml(args.yaml_file, yaml_data):
            yaml.dump(yaml_data, sys.stdout)
        else:
            json.dump(Parsers.jsonify_yaml_data(yaml_data), sys.stdout)
    else:
        save_to_file(args, log, yaml, yaml_data, backup_file)

def _try_load_input_file(args, log, yaml, change_path, new_value):
    """Attempt to load the input data file or abend on error."""
    (yaml_data, doc_loaded) = Parsers.get_yaml_data(yaml, log, args.yaml_file)
    if not doc_loaded:
        # An error message has already been logged
        sys.exit(1)
    elif yaml_data is None:
        yaml_data = Nodes.build_next_node(change_path, 0, new_value)
    return yaml_data

def _delete_nodes(log, processor, delete_nodes) -> None:
    """Recursively delete specified nodes."""
    try:
        processor.delete_gathered_nodes(delete_nodes)
    except YAMLPathException as ex:
        if "delete the entire document" in ex.user_message:
            log.critical(
                "Refusing to delete the entire document!  Ensure the source"
                " document is YAML, JSON, or compatible and --change|-g is"
                " non-empty and not the document root.", 1)

def _get_nodes(log, processor, yaml_path, **kwargs):
    """Gather requested nodes."""
    must_exist = kwargs.pop("must_exist", False)
    default_value = kwargs.pop("default_value", " ")
    ignore_fail = kwargs.pop("ignore_fail", False)
    gathered_nodes = []

    try:
        for node_coordinate in processor.get_nodes(
                yaml_path, mustexist=must_exist,
                default_value=default_value):
            log.debug(
                "Got node from {}:".format(yaml_path),
                data=node_coordinate, prefix="yaml_set::_get_nodes:  ")
            gathered_nodes.append(node_coordinate)
    except YAMLPathException as ex:
        if ignore_fail:
            log.debug(
                "Ignoring failure to gather nodes due to:  {}".format(ex),
                prefix="yaml_set::_get_nodes:  ")
            return []
        log.critical(ex, 1)

    log.debug(
        "Collected nodes:", data=gathered_nodes,
        prefix="yaml_set::_get_nodes:  ")

    return gathered_nodes

def _alias_nodes(
    log, processor, assign_to_nodes, anchor_path, anchor_name
):
    """Assign YAML Aliases to the target nodes."""
    try:
        processor.alias_gathered_nodes(
            assign_to_nodes, anchor_path, anchor_name=anchor_name)
    except YAMLPathException as ex:
        log.critical(ex, 1)

# pylint: disable=locally-disabled,too-many-arguments
def _ymk_nodes(
    log, processor, assign_to_nodes, anchor_path, anchor_name, target_path
):
    """Assign YAML Aliases to the target nodes."""
    try:
        processor.ymk_gathered_nodes(
            assign_to_nodes, anchor_path, target_path, anchor_name=anchor_name)
    except YAMLPathException as ex:
        log.critical(ex, 1)

# pylint: disable=locally-disabled,too-many-locals,too-many-branches,too-many-statements
def main():
    """Main code."""
    args = processcli()
    log = ConsolePrinter(args)
    validateargs(args, log)
    change_path = YAMLPath(args.change, pathsep=args.pathsep)
    must_exist=args.mustexist or args.saveto

    # Obtain the replacement value
    consumed_stdin = False
    new_value = None
    has_new_value = False
    if args.value or args.value == "":
        new_value = args.value
        has_new_value = True
    elif args.stdin:
        new_value = ''.join(sys.stdin.readlines())
        consumed_stdin = True
        has_new_value = True
    elif args.file:
        with open(args.file, 'r', encoding='utf-8') as fhnd:
            new_value = fhnd.read().rstrip()
        has_new_value = True
    elif args.null:
        new_value = None
        has_new_value = True
    elif args.random is not None:
        new_value = ''.join(
            secrets.choice(args.random_from) for _ in range(args.random)
        )
        has_new_value = True

    # Prep the YAML parser
    yaml = Parsers.get_yaml_editor()

    # Attempt to open the YAML file; check for parsing errors
    if args.yaml_file:
        yaml_data = _try_load_input_file(
            args, log, yaml, change_path, new_value)
        if args.yaml_file.strip() == '-':
            consumed_stdin = True

    # Check for a waiting STDIN document
    if (not consumed_stdin
        and not args.yaml_file
        and not args.nostdin
        and not sys.stdin.isatty()
    ):
        args.yaml_file = "-"
        yaml_data = _try_load_input_file(
            args, log, yaml, change_path, new_value)

    # Load the present nodes at the specified YAML Path
    processor = EYAMLProcessor(
        log, yaml_data, binary=args.eyaml,
        publickey=args.publickey, privatekey=args.privatekey)
    ignore_fail = not must_exist
    change_node_coordinates = _get_nodes(
        log, processor, change_path, must_exist=True, ignore_fail=ignore_fail,
        default_value=("" if new_value else " "))

    old_format = YAMLValueFormats.DEFAULT
    if len(change_node_coordinates) == 1:
        # When there is exactly one result, its old format can be known.  This
        # is necessary to retain whether the replacement value should be
        # represented later as a multi-line string when the new value is to be
        # encrypted.
        old_format = YAMLValueFormats.from_node(
            change_node_coordinates[0].node)

    # Check the value(s), if desired
    if args.check:
        for node_coordinate in change_node_coordinates:
            if processor.is_eyaml_value(node_coordinate.node):
                # Sanity check:  If either --publickey or --privatekey were set
                # then they must both be set in order to decrypt this value.
                # This is enforced only when the value must be decrypted due to
                # a --check request.
                if (
                        (args.publickey and not args.privatekey)
                        or (args.privatekey and not args.publickey)
                ):
                    log.error(
                        "Neither or both private and public EYAML keys must be"
                        + " set when --check is required to decrypt the old"
                        + " value.")
                    sys.exit(1)

                try:
                    check_value = processor.decrypt_eyaml(node_coordinate.node)
                except EYAMLCommandException as ex:
                    log.critical(ex, 1)
            else:
                check_value = node_coordinate.node

            if not args.check == check_value:
                log.critical(
                    '"{}" does not match the check value.'
                    .format(args.check),
                    20
                )

    # Save the old value, if desired and possible
    if args.saveto:
        # Only one can be saved; otherwise it is impossible to meaningfully
        # convey to the end-user from exactly which other YAML node each saved
        # value came.
        if len(change_node_coordinates) > 1:
            log.critical(
                "It is impossible to meaningly save more than one matched"
                + " value.  Please omit --saveto or set --change to affect"
                + " exactly one value.", 1)

        saveto_path = YAMLPath(args.saveto, pathsep=args.pathsep)
        log.verbose("Saving the old value to {}.".format(saveto_path))

        # Folded EYAML values have their embedded newlines converted to spaces
        # when read.  As such, writing them back out breaks their original
        # format, despite being properly typed.  To restore the original
        # written form, reverse the conversion, here.
        old_value = change_node_coordinates[0].node
        if (
                (old_format is YAMLValueFormats.FOLDED
                 or old_format is YAMLValueFormats.LITERAL
                )
                and EYAMLProcessor.is_eyaml_value(old_value)
        ):
            old_value = old_value.replace(" ", "\n")

        try:
            processor.set_value(
                saveto_path, Nodes.clone_node(old_value),
                value_format=old_format, tag=args.tag)
        except YAMLPathException as ex:
            log.critical(ex, 1)

    # Set the requested value
    log.verbose("Applying changes to {}.".format(change_path))
    if args.delete:
        # Destroy the collected nodes (from their parents) in the reverse order
        # they were discovered.  This is necessary lest Array elements be
        # improperly handled, leading to unwanted data loss.
        _delete_nodes(log, processor, change_node_coordinates)
    elif args.aliasof:
        # Assign the change nodes as Aliases of whatever --aliasof points to
        _alias_nodes(
            log, processor, change_node_coordinates, args.aliasof, args.anchor)
    elif args.mergekey:
        # Assign to the change nodes a YAML Merge Key to whatever --mergekey
        # points to.
        _ymk_nodes(
            log, processor, change_node_coordinates, args.mergekey,
            args.anchor, change_path)
    elif args.eyamlcrypt:
        # If the user hasn't specified a format, use the same format as the
        # value being replaced, if known.
        format_type = YAMLValueFormats.from_str(args.format)
        if format_type is YAMLValueFormats.DEFAULT:
            format_type = old_format

        output_type = EYAMLOutputFormats.STRING
        if format_type in [YAMLValueFormats.FOLDED, YAMLValueFormats.LITERAL]:
            output_type = EYAMLOutputFormats.BLOCK

        try:
            processor.set_eyaml_value(
                change_path, new_value, output=output_type, mustexist=False)
        except EYAMLCommandException as ex:
            log.critical(ex, 2)
    elif has_new_value:
        try:
            processor.set_value(
                change_path, new_value, value_format=args.format,
                mustexist=must_exist, tag=args.tag)
        except YAMLPathException as ex:
            log.critical(ex, 1)
    elif args.tag:
        processor.tag_gathered_nodes(change_node_coordinates, args.tag)

    # Write out the result
    write_output_document(args, log, yaml, yaml_data)

if __name__ == "__main__":
    main()  # pragma: no cover



import pytest

from yamlpath.merger.enums.outputdoctypes import (
	OutputDocTypes)


class Test_merger_enums_outputdoctypes():
	"""Tests for the OutputDocTypes enumeration."""

	def test_get_names(self):
		assert OutputDocTypes.get_names() == [
			"AUTO",
			"JSON",
			"YAML",
		]

	def test_get_choices(self):
		assert OutputDocTypes.get_choices() == [
			"auto",
			"json",
			"yaml",
		]

	@pytest.mark.parametrize("input,output", [
		("AUTO", OutputDocTypes.AUTO),
		("JSON", OutputDocTypes.JSON),
		("YAML", OutputDocTypes.YAML),
	])
	def test_from_str(self, input, output):
		assert output == OutputDocTypes.from_str(input)

	def test_from_str_nameerror(self):
		with pytest.raises(NameError):
			OutputDocTypes.from_str("NO SUCH NAME")



import pytest

from yamlpath.enums import PathSearchKeywords
from yamlpath.path import SearchKeywordTerms

class Test_path_SearchKeywordTerms():
    """Tests for the SearchKeywordTerms class."""

    @pytest.mark.parametrize("invert,keyword,parameters,output", [
        (True, PathSearchKeywords.HAS_CHILD, "abc", "[!has_child(abc)]"),
        (False, PathSearchKeywords.HAS_CHILD, "abc", "[has_child(abc)]"),
        (False, PathSearchKeywords.HAS_CHILD, "abc\\,def", "[has_child(abc\\,def)]"),
        (False, PathSearchKeywords.HAS_CHILD, "abc, def", "[has_child(abc, def)]"),
        (False, PathSearchKeywords.HAS_CHILD, "abc,' def'", "[has_child(abc,' def')]"),
    ])
    def test_str(self, invert, keyword, parameters, output):
        assert output == str(SearchKeywordTerms(invert, keyword, parameters))

    @pytest.mark.parametrize("parameters,output", [
        ("abc", ["abc"]),
        ("abc\\,def", ["abc,def"]),
        ("abc, def", ["abc", "def"]),
        ("abc,' def'", ["abc", " def"]),
        ("1,'1', 1, '1', 1 , ' 1', '1 ', ' 1 '", ["1", "1", "1", "1", "1", " 1", "1 ", " 1 "]),
        ("true, False,'True','false'", ["true", "False", "True", "false"]),
        ("'',,\"\", '', ,,\"\\'\",'\\\"'", ["", "", "", "", "", "", "'", "\""]),
        ("'And then, she said, \"Quote!\"'", ["And then, she said, \"Quote!\""]),
        (None, []),
    ])
    def test_parameter_parsing(self, parameters, output):
        skt = SearchKeywordTerms(False, PathSearchKeywords.HAS_CHILD, parameters)
        assert output == skt.parameters

    @pytest.mark.parametrize("parameters", [
        ("','a'"),
        ("a,\"b,"),
    ])
    def test_unmatched_demarcation(self, parameters):
        skt = SearchKeywordTerms(False, PathSearchKeywords.HAS_CHILD, parameters)
        with pytest.raises(ValueError) as ex:
            parmlist = skt.parameters
        assert -1 < str(ex.value).find("one or more unmatched demarcation symbol")



import pytest

from tests.conftest import create_temp_yaml_file


class Test_yaml_get():
    """Tests for the yaml-get command-line interface."""
    command = "yaml-get"

    def test_no_options(self, script_runner):
        result = script_runner.run(self.command, "--nostdin")
        assert not result.success, result.stderr
        assert "the following arguments are required: -p/--query" in result.stderr

    def test_no_input_file(self, script_runner):
        result = script_runner.run(self.command, "--nostdin", "--query='/test'")
        assert not result.success, result.stderr
        assert "YAML_FILE must be set or be read from STDIN" in result.stderr

    def test_bad_input_file(self, script_runner):
        result = script_runner.run(self.command, "--query='/test'", "no-such-file")
        assert not result.success, result.stderr
        assert "File not found:" in result.stderr

    def test_no_query(self, script_runner, tmp_path_factory):
        content = """---
        no: ''
        """
        yaml_file = create_temp_yaml_file(tmp_path_factory, content)
        result = script_runner.run(self.command, yaml_file)
        assert not result.success, result.stderr
        assert "the following arguments are required: -p/--query" in result.stderr

    def test_bad_privatekey(self, script_runner, tmp_path_factory):
        content = """---
        no: ''
        """
        yaml_file = create_temp_yaml_file(tmp_path_factory, content)
        result = script_runner.run(self.command, "--query=aliases", "--privatekey=no-such-file", yaml_file)
        assert not result.success, result.stderr
        assert "EYAML private key is not a readable file" in result.stderr

    def test_bad_publickey(self, script_runner, tmp_path_factory):
        content = """---
        no: ''
        """
        yaml_file = create_temp_yaml_file(tmp_path_factory, content)
        result = script_runner.run(self.command, "--query=aliases", "--publickey=no-such-file", yaml_file)
        assert not result.success, result.stderr
        assert "EYAML public key is not a readable file" in result.stderr

    def test_yaml_parsing_error(self, script_runner, imparsible_yaml_file):
        result = script_runner.run(self.command, "--query=/", imparsible_yaml_file)
        assert not result.success, result.stderr
        assert "YAML parsing error" in result.stderr

    def test_yaml_syntax_error(self, script_runner, badsyntax_yaml_file):
        result = script_runner.run(self.command, "--query=/", badsyntax_yaml_file)
        assert not result.success, result.stderr
        assert "YAML syntax error" in result.stderr

    def test_yaml_composition_error(self, script_runner, badcmp_yaml_file):
        result = script_runner.run(self.command, "--query=/", badcmp_yaml_file)
        assert not result.success, result.stderr
        assert "YAML composition error" in result.stderr

    def test_bad_yaml_path(self, script_runner, tmp_path_factory):
        content = """---
        aliases:
          - &plainScalar Plain scalar string
        """
        yaml_file = create_temp_yaml_file(tmp_path_factory, content)
        result = script_runner.run(self.command, "--query=aliases[1]", yaml_file)
        assert not result.success, result.stderr
        assert "Required YAML Path does not match any nodes" in result.stderr

    def test_bad_eyaml_value(self, script_runner, tmp_path_factory):
        content = """---
        aliases:
          - &encryptedScalar >
            ENC[PKCS7,MIIx...broken-on-purpose...==]
        """
        yaml_file = create_temp_yaml_file(tmp_path_factory, content)
        result = script_runner.run(
            self.command,
            "--query=aliases[&encryptedScalar]",
            "--eyaml=/does/not/exist-on-most/systems",
            yaml_file
        )
        assert not result.success, result.stderr
        assert "No accessible eyaml command" in result.stderr

    def test_recursive_yaml_anchor(self, script_runner, tmp_path_factory):
        content = """--- &recursive_this
hash:
  recursive_key: *recursive_this
"""
        yaml_file = create_temp_yaml_file(tmp_path_factory, content)
        result = script_runner.run(
            self.command,
            "--query=/hash",
            yaml_file
        )
        assert not result.success, result.stderr
        assert "contains an infinitely recursing" in result.stderr

    def test_query_anchor(self, script_runner, tmp_path_factory):
        content = """---
        aliases:
          - &plainScalar Plain scalar string
        """
        yaml_file = create_temp_yaml_file(tmp_path_factory, content)
        result = script_runner.run(self.command, "--query=aliases[&plainScalar]", yaml_file)
        assert result.success, result.stderr
        assert "Plain scalar string" in result.stdout

    def test_query_list(self, script_runner, tmp_path_factory):
        content = """---
        aliases:
          - &plainScalar Plain scalar string
        """
        yaml_file = create_temp_yaml_file(tmp_path_factory, content)
        result = script_runner.run(self.command, "--query=aliases", yaml_file)
        assert result.success, result.stderr
        assert '["Plain scalar string"]' in result.stdout

    def test_query_doc_from_stdin(
        self, script_runner, tmp_path_factory
    ):
        import subprocess

        yaml_file = """---
hash:
  lhs_exclusive: LHS exclusive
  merge_target: LHS original value
"""

        result = subprocess.run(
            [self.command
            , "--query=/hash/lhs_exclusive"
            , "-"]
            , stdout=subprocess.PIPE
            , input=yaml_file
            , universal_newlines=True
        )

        assert 0 == result.returncode, result.stderr
        assert "LHS exclusive\n" == result.stdout

    def test_get_every_data_type(self, script_runner, tmp_path_factory):
        # Contributed by https://github.com/AndydeCleyre
        content = """---
intthing: 6
floatthing: 6.8
yesthing: yes
nothing: no
truething: true
falsething: false
nullthing: null
nothingthing:
emptystring: ""
nullstring: "null"
datething: 2022-09-23
timestampthing: 2022-09-24T14:13:12-7:30
        """

        # Note that true nulls are translated as "\x00" (hexadecimal NULL
        # control-characters).
        results = ["6", "6.8", "yes", "no", "True", "False", "\x00", "\x00", "", "null", "2022-09-23", "2022-09-24T14:13:12-07:30"]

        yaml_file = create_temp_yaml_file(tmp_path_factory, content)
        result = script_runner.run(self.command, "--query=*", yaml_file)
        assert result.success, result.stderr

        match_index = 0
        for line in result.stdout.splitlines():
            assert line == results[match_index]
            match_index += 1

    def test_get_only_aoh_nodes_without_named_child(self, script_runner, tmp_path_factory):
        content = """---
items:
  - - alpha
    - bravo
    - charlie
  - - alpha
    - charlie
    - delta
  - - alpha
    - bravo
    - delta
  - - bravo
    - charlie
    - delta
"""
        results = [
            "alpha",
            "charlie",
            "delta"
        ]

        yaml_file = create_temp_yaml_file(tmp_path_factory, content)
        result = script_runner.run(self.command, "--query=/items/*[!has_child(bravo)]*", yaml_file)
        assert result.success, result.stderr

        match_index = 0
        for line in result.stdout.splitlines():
            assert line == results[match_index]
            match_index += 1

    def test_get_only_aoh_nodes_with_named_child(self, script_runner, tmp_path_factory):
        content = """---
products:
  - name: something
    price: 0.99
    weight: 0.75
    recalled: false
  - name: other
    price: 9.99
    weight: 2.25
    dimensions:
      width: 1
      height: 1
      depth: 1
  - name: moar
    weight: 100
    dimensions:
      width: 100
      height: 100
      depth: 100
  - name: less
    price: 5
    dimensions:
      width: 5
      height: 5
  - name: bad
    price: 0
    weight: 4
    dimensions:
      width: 13
      height: 4
      depth: 7
    recalled: true
"""
        results = [
            "something",
            "bad",
        ]

        yaml_file = create_temp_yaml_file(tmp_path_factory, content)
        result = script_runner.run(self.command, "--query=/products[has_child(recalled)]/name", yaml_file)
        assert result.success, result.stderr

        match_index = 0
        for line in result.stdout.splitlines():
            assert line == results[match_index]
            match_index += 1

    @pytest.mark.parametrize("query,output", [
        ("/items/*[!has_child(bravo)][2][parent(0)]", ['delta']),
        ("/items/*[!has_child(bravo)][2][parent()]", ['["alpha", "charlie", "delta"]']),
        ("/items/*[!has_child(bravo)][2][parent(2)]", ['[["alpha", "bravo", "charlie"], ["alpha", "charlie", "delta"], ["alpha", "bravo", "delta"], ["bravo", "charlie", "delta"]]']),
        ("/prices_hash/*[has_child(price)][name()]", ['doohickey', 'whatchamacallit', 'widget']),
        ("/prices_hash/*[!has_child(price)][name()]", ['unknown']),
    ])
    def test_get_parent_nodes(self, script_runner, tmp_path_factory, query, output):
        content = """---
items:
  - - alpha
    - bravo
    - charlie
  - - alpha
    - charlie
    - delta
  - - alpha
    - bravo
    - delta
  - - bravo
    - charlie
    - delta

prices_hash:
  doohickey:
    price: 4.99
  whatchamacallit:
    price: 9.95
  widget:
    price: 0.98
  unknown:
"""

        yaml_file = create_temp_yaml_file(tmp_path_factory, content)
        result = script_runner.run(self.command, "--query={}".format(query), yaml_file)
        assert result.success, result.stderr

        match_index = 0
        for line in result.stdout.splitlines():
            assert line == output[match_index]
            match_index += 1

    @pytest.mark.parametrize("query,output", [
        ("svcs.*[name()]", ['coolserver', 'logsender']),
        ("svcs.*[enabled=false][parent()][name()]", ['logsender']),
        ("svcs[name()]", ['svcs']),
        ("indexes[.^Item][name()]", ['0', '1', '3']),
    ])
    def test_get_node_names(self, script_runner, tmp_path_factory, query, output):
        # Contributed by https://github.com/AndydeCleyre
        content = """---
svcs:
  coolserver:
    enabled: true
    exec: ./coolserver.py
  logsender:
    enabled: false
    exec: remote_syslog -D
indexes:
  - Item 1
  - Item 2
  - Disabled 3
  - Item 4
"""

        yaml_file = create_temp_yaml_file(tmp_path_factory, content)
        result = script_runner.run(self.command, "--query={}".format(query), yaml_file)
        assert result.success, result.stderr

        match_index = 0
        for line in result.stdout.splitlines():
            assert line == output[match_index]
            match_index += 1

    @pytest.mark.parametrize("query,output", [
        ("prices_aoh[max(price)].product", ["whatchamacallit"]),
        ("prices_aoh[!max(price)].price", ["4.99", "4.99", "0.98"]),
        ("/prices_hash[max(price)][name()]", ["whatchamacallit"]),
        ("/prices_hash[!max(price)][name()]", ["doohickey", "fob", "widget", "unknown"]),
        ("(prices_hash.*.price)[max()]", ["9.95"]),
        ("(prices_hash.*.price)[!max()]", ["4.99", "4.99", "0.98"]),
        ("/prices_array[max()]", ["9.95"]),
        ("/prices_array[!max()]", ["4.99", "4.99", "0.98", "\x00"]),
        ("bare[max()]", ["value"]),
        ("/bad_prices_aoh[max(price)]/product", ["fob"]),
        ("/bad_prices_aoh[!max(price)]/price", ["4.99", "9.95", "True"]),
        ("bad_prices_hash[max(price)][name()]", ["fob"]),
        ("bad_prices_hash[!max(price)][name()]", ["doohickey", "whatchamacallit", "widget", "unknown"]),
        ("(/bad_prices_hash/*/price)[max()]", ["not set"]),
        ("(/bad_prices_hash/*/price)[!max()]", ["4.99", "9.95", "True"]),
        ("bad_prices_array[max()]", ["not set"]),
        ("bad_prices_array[!max()]", ["4.99", "9.95", "0.98", "\x00"]),

        ("prices_aoh[min(price)].product", ["widget"]),
        ("prices_aoh[!min(price)].price", ["9.95", "4.99", "4.99"]),
        ("/prices_hash[min(price)][name()]", ["widget"]),
        ("/prices_hash[!min(price)][name()]", ["whatchamacallit", "doohickey", "fob", "unknown"]),
        ("(prices_hash.*.price)[min()]", ["0.98"]),
        ("(prices_hash.*.price)[!min()]", ["9.95", "4.99", "4.99"]),
        ("/prices_array[min()]", ["0.98"]),
        ("/prices_array[!min()]", ["9.95", "4.99", "4.99", "\x00"]),
        ("bare[min()]", ["value"]),
        ("/bad_prices_aoh[min(price)]/product", ["widget"]),
        ("/bad_prices_aoh[!min(price)]/price", ["not set", "9.95", "4.99"]),
        ("bad_prices_hash[min(price)][name()]", ["widget"]),
        ("bad_prices_hash[!min(price)][name()]", ["fob", "whatchamacallit", "doohickey", "unknown"]),
        ("(/bad_prices_hash/*/price)[min()]", ["True"]),
        ("(/bad_prices_hash/*/price)[!min()]", ["not set", "9.95", "4.99"]),
        ("bad_prices_array[min()]", ["0.98"]),
        ("bad_prices_array[!min()]", ["not set", "9.95", "4.99", "\x00"]),
    ])
    def test_get_min_max_nodes(self, script_runner, tmp_path_factory, query, output):
        content = """---
# Consistent Data Types
prices_aoh:
  - product: doohickey
    price: 4.99
  - product: fob
    price: 4.99
  - product: whatchamacallit
    price: 9.95
  - product: widget
    price: 0.98
  - product: unknown

prices_hash:
  doohickey:
    price: 4.99
  fob:
    price: 4.99
  whatchamacallit:
    price: 9.95
  widget:
    price: 0.98
  unknown:

prices_array:
  - 4.99
  - 4.99
  - 9.95
  - 0.98
  - null

# Inconsistent Data Types
bare: value

bad_prices_aoh:
  - product: doohickey
    price: 4.99
  - product: fob
    price: not set
  - product: whatchamacallit
    price: 9.95
  - product: widget
    price: true
  - product: unknown

bad_prices_hash:
  doohickey:
    price: 4.99
  fob:
    price: not set
  whatchamacallit:
    price: 9.95
  widget:
    price: true
  unknown:

bad_prices_array:
  - 4.99
  - not set
  - 9.95
  - 0.98
  - null
"""

        yaml_file = create_temp_yaml_file(tmp_path_factory, content)
        result = script_runner.run(self.command, "--query={}".format(query), yaml_file)
        assert result.success, result.stderr

        match_index = 0
        for line in result.stdout.splitlines():
            assert line == output[match_index]
            match_index += 1



"""
Implements an EYAML-capable version of YAML Path.

Copyright 2018, 2019, 2020 William W. Kimball, Jr. MBA MSIS
"""
import re
from subprocess import run, PIPE, CalledProcessError
from os import access, sep, X_OK
from shutil import which
from typing import Any, Generator, List, Optional, Union

from ruamel.yaml.comments import CommentedSeq, CommentedMap

from yamlpath import YAMLPath
from yamlpath.common import Anchors
from yamlpath.eyaml.enums import EYAMLOutputFormats
from yamlpath.enums import YAMLValueFormats, PathSeperators
from yamlpath.eyaml.exceptions import EYAMLCommandException
from yamlpath.wrappers import ConsolePrinter, NodeCoords
from yamlpath import Processor


class EYAMLProcessor(Processor):
    """Extend Processor to understand EYAML content."""

    def __init__(
        self, logger: ConsolePrinter, data: Any, **kwargs: Optional[str]
    ) -> None:
        """
        Instantiate an EYAMLProcessor.

        Parameters:
        1. logger (ConsolePrinter) Instance of ConsolePrinter or subclass
        2. data (Any) Parsed YAML data

        Keyword Arguments:
        * binary (str) The external eyaml command to use when performing
            data encryption or decryption; if no path is provided, the
            command will be sought on the system PATH.  Defaut="eyaml"
        * publickey (Optional[str]) Fully-qualified path to the public key
            for use with data encryption
        * privatekey (Optional[str]) Fully-qualified path to the public key
            for use with data decryption

        Returns:  N/A

        Raises:  N/A
        """
        self.eyaml: str = str(kwargs.pop("binary", "eyaml"))
        self.publickey: Optional[str] = kwargs.pop("publickey", None)
        self.privatekey: Optional[str] = kwargs.pop("privatekey", None)
        super().__init__(logger, data)

    # pylint: disable=locally-disabled,too-many-branches
    def _find_eyaml_paths(
        self, data: Any, build_path: YAMLPath
    ) -> Generator[YAMLPath, None, None]:
        """
        Find every encrypted value and report each as a YAML Path.

        Recursively generates a set of stringified YAML Paths, each entry
        leading to an EYAML value within the evaluated YAML data.

        Parameters:
        1. data (Any) The parsed YAML data to process
        2. build_path (YAMLPath) A YAML Path under construction

        Returns:  (Generator[Path, None, None]) each YAML Path entry as they
            are discovered

        Raises:  N/A
        """
        if isinstance(data, CommentedSeq):
            for idx, ele in enumerate(data):
                node_anchor = Anchors.get_node_anchor(ele)
                if node_anchor is not None:
                    escaped_section = YAMLPath.escape_path_section(
                        node_anchor, PathSeperators.DOT)
                    tmp_path_segment = f"[&{escaped_section}]"
                else:
                    tmp_path_segment = f"[{idx}]"

                tmp_path = build_path + tmp_path_segment
                if self.is_eyaml_value(ele):
                    yield tmp_path
                else:
                    for subpath in self._find_eyaml_paths(ele, tmp_path):
                        yield subpath

        elif isinstance(data, CommentedMap):
            for key, val in data.non_merged_items():
                tmp_path = build_path + YAMLPath.escape_path_section(
                    key, PathSeperators.DOT)
                if self.is_eyaml_value(val):
                    yield tmp_path
                else:
                    for subpath in self._find_eyaml_paths(val, tmp_path):
                        yield subpath

    def find_eyaml_paths(self) -> Generator[YAMLPath, None, None]:
        """
        Find every encrypted value and reports its YAML Path.

        Parameters:  N/A

        Returns:  (Generator[Path, None, None]) each YAML Path entry as they
            are discovered

        Raises:  N/A
        """
        # Initiate the scan from the data root
        for path in self._find_eyaml_paths(self.data, YAMLPath()):
            yield path

    def decrypt_eyaml(
        self, value: Union[str, list, NodeCoords]
    ) -> Union[str, list]:
        """
        Decrypt an EYAML value.

        Parameters:
        1. value (str) The EYAML value to decrypt

        Returns:  (str) The decrypted value or the original value if it was not
            actually encrypted.

        Raises:
        - `EYAMLCommandException` when the eyaml binary cannot be utilized
        """
        if isinstance(value, NodeCoords):
            value = self.decrypt_eyaml(value.node)

        # Recursively decrypt lists
        if isinstance(value, list):
            decvals: list = []
            for encval in value:
                decvals.append(self.decrypt_eyaml(encval))
            return decvals

        if not self.is_eyaml_value(value):
            return value

        if not self._can_run_eyaml():
            raise EYAMLCommandException("No accessible eyaml command.")

        cmd: List[str] = [
            self.eyaml,
            'decrypt',
            '--quiet',
            '--stdin'
        ]
        if self.publickey:
            cmd.append(f"--pkcs7-public-key={self.publickey}")
        if self.privatekey:
            cmd.append(f"--pkcs7-private-key={self.privatekey}")

        cleanval: str = str(value).replace("\n", "").replace(" ", "").rstrip()
        bval: bytes = cleanval.encode("ascii")
        self.logger.debug(
            f"About to execute {' '.join(cmd)} against:\n{cleanval}",
            prefix="EYAMLPath::decrypt_eyaml:  "
        )

        try:
            # self.eyaml is untrusted, so shell must always be False and
            # all parameters must be supplied via a List.
            retval: str = run(
                cmd,
                stdout=PIPE,
                input=bval,
                check=True,
                shell=False
            ).stdout.decode('ascii').rstrip()
        except CalledProcessError as ex:
            raise EYAMLCommandException(
                f"The {self.eyaml} command cannot be run due to exit code:"
                f"  {ex.returncode}"
            ) from ex

        # Check for bad decryptions
        self.logger.debug(
            f"EYAMLPath::decrypt_eyaml:  Decrypted result:  {retval}"
        )
        if not retval or retval == cleanval:
            raise EYAMLCommandException(
                "Unable to decrypt value!  Please verify you are using the"
                " correct old EYAML keys and the value is not corrupt:"
                "    {cleanval}"
            )

        return retval

    def encrypt_eyaml(
        self, value: str,
        output: EYAMLOutputFormats = EYAMLOutputFormats.STRING
    ) -> str:
        """
        Encrypt a value via EYAML.

        Parameters:
        1. value (str) the value to encrypt
        2. output (EYAMLOutputFormats) the output format of the encryption

        Returns:  (str) The encrypted result or the original value if it was
            already an EYAML encryption.

        Raises:
        - `EYAMLCommandException` when the eyaml binary cannot be utilized.
        """
        if self.is_eyaml_value(value):
            return value

        if not self._can_run_eyaml():
            raise EYAMLCommandException(
                f"The eyaml binary is not executable at:  {self.eyaml}"
            )

        cmd: List[str] = [
            self.eyaml,
            'encrypt',
            '--quiet',
            '--stdin',
            f"--output={output}"
        ]
        if self.publickey:
            cmd.append(f"--pkcs7-public-key={self.publickey}")
        if self.privatekey:
            cmd.append(f"--pkcs7-private-key={self.privatekey}")

        self.logger.debug(
            f"EYAMLPath::encrypt_eyaml:  About to execute:  {' '.join(cmd)}"
        )
        bval: bytes = value.encode("ascii")

        try:
            # self.eyaml is untrusted, so shell must always be False and
            # all parameters must be supplied via a List.
            retval: str = (
                run(cmd, stdout=PIPE, input=bval, check=True, shell=False)
                .stdout
                .decode("ascii")
                .rstrip()
            )
        except CalledProcessError as ex:
            raise EYAMLCommandException(
                f"The {self.eyaml} command cannot be run due to exit code:"
                f"  {ex.returncode}"
            ) from ex

        # While exceedingly rare and difficult to test for, it is possible
        # for custom eyaml commands to produce no output.  This is a critical
        # error in every conceivable case but pycov will never get a test
        # that works multi-platform.  So, ignore covering this case.
        if not retval: # pragma: no cover
            raise EYAMLCommandException(
                f"The {self.eyaml} command was unable to encrypt your value."
                "  Please verify this process can run that command and read"
                " your EYAML keys."
            )

        if output is EYAMLOutputFormats.BLOCK:
            fixval: str = re.sub(r" ", "", retval.strip())
            retval = re.sub(r"\r\n", " ", fixval) + "\n"

        self.logger.debug(
            f"Encrypted result:\n{retval}",
            prefix="EYAMLPath::encrypt_eyaml:  "
        )
        return retval

    def set_eyaml_value(
        self, yaml_path: YAMLPath, value: str,
        output: EYAMLOutputFormats = EYAMLOutputFormats.STRING,
        mustexist: bool = False
    ) -> None:
        """
        Encrypt and store a value where specified via YAML Path.

        Parameters:
        1. yaml_path (Path) The YAML Path specifying which nodes are to
           receive the encrypted value
        2. value (any) The value to encrypt
        3. output (EYAMLOutputFormats) the output format of the encryption
        4. mustexist (bool) Indicates whether YAML Path must
           specify a pre-existing node

        Returns:  N/A

        Raises:
        - `YAMLPathException` when YAML Path is invalid
        """
        self.logger.verbose(
            f"Encrypting value(s) for {yaml_path} using {output} format."
        )
        encval: str = self.encrypt_eyaml(value, output)
        emit_format: YAMLValueFormats = YAMLValueFormats.FOLDED
        if output is EYAMLOutputFormats.STRING:
            emit_format = YAMLValueFormats.DEFAULT

        self.set_value(
            yaml_path,
            encval,
            mustexist=mustexist,
            value_format=emit_format
        )

    def get_eyaml_values(
        self, yaml_path: YAMLPath, mustexist: bool = False,
        default_value: str = ""
    ) -> Generator[Union[str, list], None, None]:
        """
        Retrieve and decrypt all EYAML nodes identified via a YAML Path.

        Parameters:
        1. yaml_path (Path) The YAML Path specifying which nodes to decrypt
        2. mustexist (bool) Indicates whether YAML Path must specify a
           pre-existing node; when False, the node will be created when
           missing
        3. default_value (str) The default value to add to the YAML data
           when `mustexist=False` and yaml_path points to a non-existent
           node

        Returns:  (str) The decrypted value or `default_value` when YAML Path
            specifies a non-existant node

        Raises:
        - `YAMLPathException` when YAML Path is invalid
        """
        self.logger.verbose(f"Decrypting value(s) at {yaml_path}.")
        for node in self.get_nodes(yaml_path, mustexist=mustexist,
                                   default_value=default_value):
            plain_text: Union[str, list] = self.decrypt_eyaml(node.node)
            yield plain_text

    def _can_run_eyaml(self) -> bool:
        """
        Indicate whether this instance is capable of running the eyaml binary.

        Parameters:  N/A

        Returns:  (bool) True when the present eyaml property indicates an
            executable; False, otherwise

        Raises:  N/A
        """
        binary: Optional[str] = EYAMLProcessor.get_eyaml_executable(self.eyaml)
        if binary is None:
            return False
        self.eyaml = binary
        return True

    @staticmethod
    def get_eyaml_executable(binary: Optional[str] = "eyaml") -> Optional[str]:
        """
        Return the full executable path to an eyaml binary.

        Returns None when it cannot be found or is not executable.

        Parameters:
        1. binary (str) The executable to test; if an absolute or relative
           path is not provided, the system PATH will be searched for a
           match to test

        Returns: (str) None or the executable eyaml binary path

        Raises:  N/A
        """
        if binary is None or not binary:
            return None

        if binary.find(sep) < 0:
            binary = which(binary)
            if binary is None:
                return None
            binary = str(binary)

        if access(binary, X_OK):
            return binary
        return None

    @staticmethod
    def is_eyaml_value(value: str) -> bool:
        """
        Indicate whether a value is EYAML-encrypted.

        Parameters:
        1. value (any) The value to check

        Returns:  (bool) True when the value is encrypted; False, otherwise

        Raises:  N/A
        """
        if not isinstance(value, str):
            return False
        return value.replace("\n", "").replace(" ", "").startswith("ENC[")



import pytest

from yamlpath.merger.enums.anchorconflictresolutions import (
	AnchorConflictResolutions)


class Test_merger_enum_anchorconflictresolutions():
	"""Tests for the AnchorConflictResolutions enumeration."""

	def test_get_names(self):
		assert AnchorConflictResolutions.get_names() == [
			"STOP",
			"LEFT",
			"RIGHT",
			"RENAME",
		]

	def test_get_choices(self):
		assert AnchorConflictResolutions.get_choices() == [
			"left",
			"rename",
			"right",
			"stop",
		]

	@pytest.mark.parametrize("input,output", [
		("STOP", AnchorConflictResolutions.STOP),
		("LEFT", AnchorConflictResolutions.LEFT),
		("RIGHT", AnchorConflictResolutions.RIGHT),
		("RENAME", AnchorConflictResolutions.RENAME),
	])
	def test_from_str(self, input, output):
		assert output == AnchorConflictResolutions.from_str(input)

	def test_from_str_nameerror(self):
		with pytest.raises(NameError):
			AnchorConflictResolutions.from_str("NO SUCH NAME")



"""
Implement Searches, a static library of generally-useful code for searching.

Copyright 2020 William W. Kimball, Jr. MBA MSIS
"""
import re
from typing import Any, List

from yamlpath.enums import (
    AnchorMatches,
    PathSearchMethods,
)
from yamlpath.common import Anchors, Nodes
from yamlpath.types import PathAttributes
from yamlpath.path import SearchTerms


class Searches:
    """Helper methods for common data searching operations."""

    @staticmethod
    # pylint: disable=too-many-branches,too-many-statements
    def search_matches(
        method: PathSearchMethods, needle: str, haystack: Any
    ) -> bool:
        """
        Perform a search comparison.

        NOTE:  For less-than, greather-than and related operations, the test is
        whether `haystack` is less/greater-than `needle`.

        Parameters:
        1. method (PathSearchMethods) The search method to employ
        2. needle (str) The value to look for.
        3. haystack (Any) The value to look in.

        Returns:  (bool) True = comparision passes; False = comparison fails.
        """
        typed_haystack = Nodes.typed_value(haystack)
        typed_needle = Nodes.typed_value(needle)
        needle_type = type(typed_needle)
        matches: bool = False

        if method is PathSearchMethods.EQUALS:
            if isinstance(typed_haystack, bool) and needle_type is bool:
                matches = typed_haystack == typed_needle
            elif isinstance(typed_haystack, int) and needle_type is int:
                matches = typed_haystack == typed_needle
            elif isinstance(typed_haystack, float) and needle_type is float:
                matches = typed_haystack == typed_needle
            else:
                matches = str(typed_haystack) == str(needle)
        elif method is PathSearchMethods.STARTS_WITH:
            matches = str(typed_haystack).startswith(needle)
        elif method is PathSearchMethods.ENDS_WITH:
            matches = str(typed_haystack).endswith(needle)
        elif method is PathSearchMethods.CONTAINS:
            matches = needle in str(typed_haystack)
        elif method is PathSearchMethods.GREATER_THAN:
            if isinstance(typed_haystack, int):
                if isinstance(typed_needle, (int, float)):
                    matches = typed_haystack > typed_needle
                else:
                    matches = False
            elif isinstance(typed_haystack, float):
                if isinstance(typed_needle, (int, float)):
                    matches = typed_haystack > typed_needle
                else:
                    matches = False
            else:
                matches = str(typed_haystack) > str(needle)
        elif method is PathSearchMethods.LESS_THAN:
            if isinstance(typed_haystack, int):
                if isinstance(typed_needle, (int, float)):
                    matches = typed_haystack < typed_needle
                else:
                    matches = False
            elif isinstance(typed_haystack, float):
                if isinstance(typed_needle, (int, float)):
                    matches = typed_haystack < typed_needle
                else:
                    matches = False
            else:
                matches = str(typed_haystack) < str(needle)
        elif method is PathSearchMethods.GREATER_THAN_OR_EQUAL:
            if isinstance(typed_haystack, int):
                if isinstance(typed_needle, (int, float)):
                    matches = typed_haystack >= typed_needle
                else:
                    matches = False
            elif isinstance(typed_haystack, float):
                if isinstance(typed_needle, (int, float)):
                    matches = typed_haystack >= typed_needle
                else:
                    matches = False
            else:
                matches = str(typed_haystack) >= str(needle)
        elif method is PathSearchMethods.LESS_THAN_OR_EQUAL:
            if isinstance(typed_haystack, int):
                if isinstance(typed_needle, (int, float)):
                    matches = typed_haystack <= typed_needle
                else:
                    matches = False
            elif isinstance(typed_haystack, float):
                if isinstance(typed_needle, (int, float)):
                    matches = typed_haystack <= typed_needle
                else:
                    matches = False
            else:
                matches = str(typed_haystack) <= str(needle)
        elif method == PathSearchMethods.REGEX:
            matcher = re.compile(needle)
            matches = matcher.search(str(typed_haystack)) is not None
        else:
            raise NotImplementedError

        return matches

    @staticmethod
    def search_anchor(
        node: Any, terms: SearchTerms, seen_anchors: List[str], **kwargs: bool
    ) -> AnchorMatches:
        """
        Indicate whether a node has an Anchor matching given search terms.

        Parameters:
        1. node (Any) The node to search (the haystack)
        2. terms (SearchTerms) The search terms (the needle)
        3. seen_anchors (List[str]) Tracks whether the present Anchor under
           evaluation is really an Alias to another node

        Keyword Arguments:
        * search_anchors (bool) User-specific preference indicating whether to
          search Anchors and/or Aliases
        * include_aliases (bool) User-specified preference indicating whether
          to include Aliases in search results

        Returns:  (AnchorMatches) The search result
        """
        anchor_name = Anchors.get_node_anchor(node)
        if anchor_name is None:
            return AnchorMatches.NO_ANCHOR

        is_alias = True
        if anchor_name not in seen_anchors:
            is_alias = False
            seen_anchors.append(anchor_name)

        search_anchors: bool = kwargs.pop("search_anchors", False)
        if not search_anchors:
            retval = AnchorMatches.UNSEARCHABLE_ANCHOR
            if is_alias:
                retval = AnchorMatches.UNSEARCHABLE_ALIAS
            return retval

        include_aliases: bool = kwargs.pop("include_aliases", False)
        if is_alias and not include_aliases:
            return AnchorMatches.ALIAS_EXCLUDED

        retval = AnchorMatches.NO_MATCH
        matches = Searches.search_matches(
            terms.method, terms.term, anchor_name)
        if ((matches and not terms.inverted)
            or (terms.inverted and not matches)
        ):
            retval = AnchorMatches.MATCH
            if is_alias:
                retval = AnchorMatches.ALIAS_INCLUDED
        return retval

    @staticmethod
    def create_searchterms_from_pathattributes(
        rhs: PathAttributes
    ) -> SearchTerms:
        """
        Convert a PathAttributes instance to a SearchTerms instance.

        Parameters:
        1. rhs (PathAttributes) PathAttributes instance to convert

        Returns:  (SearchTerms) SearchTerms extracted from `rhs`
        """
        if isinstance(rhs, SearchTerms):
            newinst: SearchTerms = SearchTerms(
                rhs.inverted, rhs.method, rhs.attribute, rhs.term
            )
            return newinst
        raise AttributeError



"""
Implement Parsers, a static library of generally-useful code for data parsers.

Copyright 2020, 2021 William W. Kimball, Jr. MBA MSIS
"""
import warnings
from sys import maxsize, stdin
from datetime import date, datetime
from typing import Any, Dict, Generator, Tuple

import ruamel.yaml # type: ignore
from ruamel.yaml import YAML
from ruamel.yaml.parser import ParserError
from ruamel.yaml.composer import ComposerError, ReusedAnchorWarning
from ruamel.yaml.constructor import ConstructorError, DuplicateKeyError
from ruamel.yaml.scanner import ScannerError
from ruamel.yaml.scalarbool import ScalarBoolean
from ruamel.yaml.scalarstring import ScalarString
from ruamel.yaml.comments import (
    CommentedMap, CommentedSet, CommentedSeq, TaggedScalar
)
# pylint: disable=wrong-import-position,ungrouped-imports
from ruamel.yaml import version_info as ryversion
if ryversion < (0, 17, 22):                   # pragma: no cover
    from yamlpath.patches.timestamp import (
        AnchoredTimeStamp,
        AnchoredDate,
    )  # type: ignore
else:                                         # pragma: no cover
    # Temporarily fool MYPY into resolving the future-case imports
    from ruamel.yaml.timestamp import TimeStamp as AnchoredTimeStamp
    AnchoredDate = AnchoredTimeStamp
    #from ruamel.yaml.timestamp import AnchoredTimeStamp
    # From whence shall come AnchoredDate?

from yamlpath.wrappers import ConsolePrinter
from yamlpath.common import Nodes

if ruamel.yaml.version_info < (0, 17, 5):                # pragma: no cover
    from yamlpath.patches.aliasstyle import MySerializer # type: ignore
    from yamlpath.patches.aliasstyle import MyEmitter    # type: ignore
# pylint: enable=wrong-import-position,ungrouped-imports


class Parsers:
    """Helper methods for common YAML/JSON/Compatible parser operations."""

    @staticmethod
    def get_yaml_editor(**kwargs: Any) -> YAML:
        """
        Build and return a generic YAML editor based on ruamel.yaml.

        Parameters:  N/A

        Keyword Arguments:
        * explicit_start (bool) True = ensure the YAML Start-of-Document marker
          (---<EOL>) is written in the output; False = remove it; default=True
        * explode_aliases (bool) True = convert all aliases (*name) and YAML
          merge operators (<<: *name) to their referenced content, removing the
          aliases and merge operators; False = maintain the references;
          default=False
        * preserve_quotes (bool) True = retain any and all quoting of keys and
          values including whatever demarcation symbol was used (" versus ');
          False = only quote values when necessary, removing unnecessary
          demarcation; default=True

        Returns (Any) The ready-for-use YAML editor.

        Raises:  N/A
        """
        explicit_start = kwargs.pop("explicit_start", True)
        explode_aliases = kwargs.pop("explode_aliases", False)
        preserve_quotes = kwargs.pop("preserve_quotes", True)

        # The ruamel.yaml class appears to be missing some typing data, so
        # these valid assignments cannot be type-checked.
        yaml = YAML()

        # Import Anthon's patch for Aliased entries in Unordered Sets per
        # https://sourceforge.net/p/ruamel-yaml/tickets/384/
        if ruamel.yaml.version_info < (0, 17, 5):  # pragma: no cover
            yaml.Serializer = MySerializer         # type: ignore
            yaml.Emitter = MyEmitter               # type: ignore

        yaml.indent(mapping=2, sequence=4, offset=2)
        yaml.explicit_start = explicit_start       # type: ignore
        yaml.preserve_quotes = preserve_quotes     # type: ignore
        yaml.width = maxsize                       # type: ignore

        if explode_aliases:
            yaml.default_flow_style = False

        return yaml

    @staticmethod
    # pylint: disable=too-many-branches,too-many-statements,too-many-locals
    def get_yaml_data(
        parser: Any, logger: ConsolePrinter, source: str, **kwargs
    ) -> Tuple[Any, bool]:
        """
        Parse YAML/Compatible data and return the ruamel.yaml object result.

        All known issues are caught and distinctively logged.

        Parameters:
        1. parser (ruamel.yaml.YAML) The YAML data parser
        2. logger (ConsolePrinter) The logging facility
        3. source (str) The source file or serialized literal to load; can be -
           for reading from STDIN (implies literal=True)

        Keyword Arguments:
        * literal (bool) `source` is literal serialized YAML data rather than a
          file-spec, so load it directly

        Returns:  Tuple[Any, bool] A tuple containing the document and its
        success/fail state.  The first field is the parsed document; will be
        None for empty documents and for documents which could not be read.
        The second field will be True when there were no errors during parsing
        and False, otherwise.
        """
        literal = kwargs.pop("literal", False)
        yaml_data = None
        data_available = True

        # This code traps errors and warnings from ruamel.yaml, substituting
        # lengthy stack-dumps with specific, meaningful feedback.  Further,
        # some warnings are treated as errors by ruamel.yaml, so these are also
        # coallesced into cleaner feedback.
        try:
            with warnings.catch_warnings():
                warnings.filterwarnings("error")
                if source == "-":
                    yaml_data = parser.load(stdin.read())
                else:
                    if literal:
                        yaml_data = parser.load(source)
                    else:
                        with open(source, 'r', encoding='utf-8') as fhnd:
                            yaml_data = parser.load(fhnd)
        except KeyboardInterrupt:
            logger.error("Aborting data load due to keyboard interrupt!")
            data_available = False
        except FileNotFoundError:
            logger.error("File not found:  {}".format(source))
            data_available = False
        except ParserError as ex:
            logger.error("YAML parsing error {}:  {}"
                        .format(str(ex.problem_mark).lstrip(), ex.problem))
            data_available = False
        except ComposerError as ex:
            logger.error("YAML composition error {}:  {}"
                        .format(str(ex.problem_mark).lstrip(), ex.problem))
            data_available = False
        except ConstructorError as ex:
            logger.error("YAML construction error {}:  {}"
                        .format(str(ex.problem_mark).lstrip(), ex.problem))
            data_available = False
        except ScannerError as ex:
            logger.error("YAML syntax error {}:  {}"
                        .format(str(ex.problem_mark).lstrip(), ex.problem))
            data_available = False
        except DuplicateKeyError as dke:
            omits = [
                "while constructing", "To suppress this", "readthedocs",
                "future releases", "the new API",
            ]
            message = str(dke).split("\n")
            newmsg = ""
            for line in message:
                line = line.strip()
                if not line:
                    continue
                write_line = True
                for omit in omits:
                    if omit in line:
                        write_line = False
                        break
                if write_line:
                    newmsg += "\n   " + line
            logger.error("Duplicate Hash key detected:  {}"
                        .format(newmsg))
            data_available = False
        except ReusedAnchorWarning as raw:
            logger.error("Duplicate YAML Anchor detected:  {}"
                        .format(
                            str(raw)
                            .replace("occurrence   ", "occurrence ")
                            .replace("\n", "\n   ")))
            data_available = False

        return (yaml_data, data_available)

    @staticmethod
    # pylint: disable=too-many-branches,too-many-statements,too-many-locals
    def get_yaml_multidoc_data(
        parser: Any, logger: ConsolePrinter, source: str, **kwargs
    ) -> Generator[Tuple[Any, bool], None, None]:
        """
        Parse YAML/Compatible multi-docs and yield each ruamel.yaml object.

        All known issues are caught and distinctively logged.

        Parameters:
        1. parser (ruamel.yaml.YAML) The YAML data parser
        2. logger (ConsolePrinter) The logging facility
        3. source (str) The source file to load; can be - for reading from
           STDIN

        Keyword Arguments:
        * literal (bool) `source` is literal serialized YAML data rather than a
          file-spec, so load it directly

        Returns:  Generator[Tuple[Any, bool], None, None] A tuple for each
        document as it is parsed.  The first field is the parsed document; will
        be None for empty documents and for documents which could not be read.
        The second field will be True when there were no errors during parsing
        and False, otherwise.
        """
        literal = kwargs.pop("literal", False)

        # This code traps errors and warnings from ruamel.yaml, substituting
        # lengthy stack-dumps with specific, meaningful feedback.  Further,
        # some warnings are treated as errors by ruamel.yaml, so these are also
        # coallesced into cleaner feedback.
        has_error = False
        try:
            with warnings.catch_warnings():
                warnings.filterwarnings("error")
                if source == "-":
                    doc_yielded = False
                    for document in parser.load_all(stdin.read()):
                        doc_yielded = True
                        logger.debug(
                            "Yielding document from {}:".format(source),
                            prefix="get_yaml_multidoc_data: ", data=document)
                        yield (document, True)

                    # The user sent a deliberately empty document via STDIN
                    if not doc_yielded:
                        yield ("", True)
                else:
                    if literal:
                        for document in parser.load_all(source):
                            yield (document, True)
                    else:
                        with open(source, 'r', encoding='utf-8') as fhnd:
                            for document in parser.load_all(fhnd):
                                logger.debug(
                                    "Yielding document from {}:"
                                    .format(source),
                                    prefix="get_yaml_multidoc_data: ",
                                    data=document)
                                yield (document, True)
        except KeyboardInterrupt:
            has_error = True
            logger.error("Aborting data load due to keyboard interrupt!")
        except FileNotFoundError:
            has_error = True
            logger.error("File not found:  {}".format(source))
        except ParserError as ex:
            has_error = True
            logger.error("YAML parsing error {}:  {}"
                        .format(str(ex.problem_mark).lstrip(), ex.problem))
        except ComposerError as ex:
            has_error = True
            logger.error("YAML composition error {}:  {}"
                        .format(str(ex.problem_mark).lstrip(), ex.problem))
        except ConstructorError as ex:
            has_error = True
            logger.error("YAML construction error {}:  {}"
                        .format(str(ex.problem_mark).lstrip(), ex.problem))
        except ScannerError as ex:
            has_error = True
            logger.error("YAML syntax error {}:  {}"
                        .format(str(ex.problem_mark).lstrip(), ex.problem))
        except DuplicateKeyError as dke:
            has_error = True
            omits = [
                "while constructing", "To suppress this", "readthedocs",
                "future releases", "the new API",
            ]
            message = str(dke).split("\n")
            newmsg = ""
            for line in message:
                line = line.strip()
                if not line:
                    continue
                write_line = True
                for omit in omits:
                    if omit in line:
                        write_line = False
                        break
                if write_line:
                    newmsg += "\n   " + line
            logger.error("Duplicate Hash key detected:  {}"
                        .format(newmsg))
        except ReusedAnchorWarning as raw:
            has_error = True
            logger.error("Duplicate YAML Anchor detected:  {}"
                        .format(
                            str(raw)
                            .replace("occurrence   ", "occurrence ")
                            .replace("\n", "\n   ")))

        if has_error:
            yield (None, False)

    @staticmethod
    def stringify_dates(data: Any) -> Any:
        """
        Recurse through a data structure, converting all dates to strings.

        The jsonify_yaml_data covers more data-types than just dates.  This
        stringify_dates method may be removed in a future version in favor of
        the more comprehensive jsonify_yaml_data method.

        This is required for JSON output, which has no serialization support
        for native date objects.
        """
        if isinstance(data, CommentedMap):
            for key, val in data.items():
                data[key] = Parsers.stringify_dates(val)
        elif isinstance(data, CommentedSeq):
            for idx, ele in enumerate(data):
                data[idx] = Parsers.stringify_dates(ele)
        elif isinstance(data, (datetime, date)):
            return data.isoformat()
        return data

    @staticmethod
    # pylint: disable=too-many-branches
    def jsonify_yaml_data(data: Any) -> Any:
        """
        Convert all non-JSON-serializable values to strings.

        This is required when writing to JSON, which has no serialization
        support for certain YAML extensions -- like tags -- and some otherwise
        native data-types, like dates.
        """
        if isinstance(data, CommentedMap):
            for i, k in [
                (idx, key) for idx, key in enumerate(data.keys())
                if isinstance(key, TaggedScalar)
            ]:
                unwrapped_key = Parsers.jsonify_yaml_data(k)
                data.insert(i, unwrapped_key, data.pop(k))

            for key, val in data.items():
                data[key] = Parsers.jsonify_yaml_data(val)
        elif isinstance(data, dict):
            for key, val in data.items():
                data[key] = Parsers.jsonify_yaml_data(val)
        elif isinstance(data, (list, CommentedSeq)):
            for idx, ele in enumerate(data):
                data[idx] = Parsers.jsonify_yaml_data(ele)
        elif isinstance(data, (set, CommentedSet)):
            json_repr: Dict[str, None] = {}
            for set_val in data:
                json_key_proto = Parsers.jsonify_yaml_data(set_val)
                json_key = (str(json_key_proto)
                            if isinstance(json_key_proto, tuple)
                            else json_key_proto)
                json_repr[json_key] = None
            data = json_repr
        elif isinstance(data, TaggedScalar):
            if data.tag.value == "!null":
                return None
            data = Parsers.jsonify_yaml_data(data.value)
        elif isinstance(data, AnchoredDate):
            data = data.date().isoformat()
        elif isinstance(data, AnchoredTimeStamp):
            data = Nodes.get_timestamp_with_tzinfo(data).isoformat()
        elif isinstance(data, (datetime, date)):
            data = data.isoformat()
        elif isinstance(data, bytes):
            data = str(data)
        elif isinstance(data, (ScalarBoolean, bool)):
            data = bool(data)
        return data

    @staticmethod
    def delete_all_comments(dom: Any) -> None:
        """
        Recursively delete all comments from a YAML document.

        See:
        https://stackoverflow.com/questions/60080325/how-to-delete-all-comments-in-ruamel-yaml/60099750#60099750

        Parameters:
        1. dom (Any) The document to strip of all comments.

        Returns:  N/A
        """
        if dom is None:
            return

        if isinstance(dom, CommentedMap):
            for key, val in dom.items():
                Parsers.delete_all_comments(key)
                Parsers.delete_all_comments(val)
        elif isinstance(dom, CommentedSeq):
            for ele in dom:
                Parsers.delete_all_comments(ele)
        try:
            # literal scalarstring might have comment associated with them
            attr = "comment" if isinstance(dom, ScalarString) \
                else ruamel.yaml.comments.Comment.attrib
            delattr(dom, attr)
        except AttributeError:
            pass

    @staticmethod
    def set_flow_style(node: Any, is_flow: bool) -> None:
        """
        Recursively apply flow|block style to a node.

        Parameters:
        1. node (Any) The node to apply flow|block style to.
        2. is_flow (bool) True=flow-style, False=block-style

        Returns:  N/A
        """
        if hasattr(node, "fa"):
            if is_flow:
                node.fa.set_flow_style()
            else:
                node.fa.set_block_style()

        if isinstance(node, CommentedMap):
            for key, val in node.non_merged_items():
                Parsers.set_flow_style(key, is_flow)
                Parsers.set_flow_style(val, is_flow)
        elif isinstance(node, CommentedSeq):
            for ele in node:
                Parsers.set_flow_style(ele, is_flow)



import pytest

from tests.conftest import create_temp_yaml_file


class Test_commands_yaml_validate():
    """Tests for the yaml-validate command-line tool."""
    command = "yaml-validate"

    def test_no_arguments(self, script_runner):
        result = script_runner.run(self.command, "--nostdin")
        assert not result.success, result.stderr
        assert "There must be at least one YAML_FILE" in result.stderr

    def test_too_many_pseudofiles(self, script_runner):
        result = script_runner.run(
            self.command
            , '-'
            , '-')
        assert not result.success, result.stderr
        assert "Only one YAML_FILE may be the - pseudo-file" in result.stderr

    def test_valid_singledoc(self, script_runner, tmp_path_factory):
        yaml_file = create_temp_yaml_file(tmp_path_factory, """---
this:
  single-document:
    is: valid
""")
        result = script_runner.run(
            self.command
            , "--nostdin"
            , yaml_file)
        assert result.success, result.stderr

    def test_invalid_singledoc(self, script_runner, tmp_path_factory):
        yaml_file = create_temp_yaml_file(tmp_path_factory, "{[}")
        result = script_runner.run(
            self.command
            , "--nostdin"
            , yaml_file)
        assert not result.success, result.stderr
        assert "  * YAML parsing error in" in result.stdout

    def test_valid_stdin_explicit(self, script_runner, tmp_path_factory):
        import subprocess
        stdin_content = "{this: {is: valid}}"
        result = subprocess.run(
            [self.command
            , "-"]
            , stdout=subprocess.PIPE
            , input=stdin_content
            , universal_newlines=True
        )
        assert 0 == result.returncode, result.stderr

    def test_valid_stdin_implicit(self, script_runner, tmp_path_factory):
        import subprocess
        stdin_content = "{this: {is: valid}}"
        result = subprocess.run(
            [self.command]
            , stdout=subprocess.PIPE
            , input=stdin_content
            , universal_newlines=True
        )
        assert 0 == result.returncode, result.stderr

    def test_invalid_stdin_explicit(self, script_runner, tmp_path_factory):
        import subprocess
        stdin_content = "{this: {is not: valid}]"
        result = subprocess.run(
            [self.command
            , "-"]
            , stdout=subprocess.PIPE
            , input=stdin_content
            , universal_newlines=True
        )
        assert 2 == result.returncode, result.stderr
        assert "  * YAML parsing error in" in result.stdout

    def test_invalid_stdin_implicit(self, script_runner, tmp_path_factory):
        import subprocess
        stdin_content = "{this: {is not: valid}]"
        result = subprocess.run(
            [self.command]
            , stdout=subprocess.PIPE
            , input=stdin_content
            , universal_newlines=True
        )
        assert 2 == result.returncode, result.stderr
        assert "  * YAML parsing error in" in result.stdout



import pytest

from yamlpath.merger.exceptions import MergeException


class Test_exceptions_mergeexception():
	def test_str_pathless(self):
		message = "Test message"
		with pytest.raises(MergeException) as ex:
			raise MergeException(message)
		assert str(ex.value) == message

	def test_str_pathfull(self):
		message = "Test message"
		path = "/test"
		with pytest.raises(MergeException) as ex:
			raise MergeException(message, path)
		assert (str(ex.value) == "{}  This issue occurred at YAML Path:  {}"
			.format(message, path))



"""
Implements the DiffActions enumeration.

Copyright 2019, 2020 William W. Kimball, Jr. MBA MSIS
"""
from enum import Enum, auto


class DiffActions(Enum):
    """
    The action taken for one document to become the next.

    `ADD`
        Add an entry

    `CHANGE`
        An entry has changed.

    `DELETE`
        Remove an entry.

    `SAME`
        The two entries are identical.
    """

    ADD = auto()
    CHANGE = auto()
    DELETE = auto()
    SAME = auto()

    def __str__(self) -> str:
        """Get the diff-like entry code for this action."""
        diff_type = ""
        if self is DiffActions.ADD:
            diff_type = "a"
        elif self is DiffActions.CHANGE:
            diff_type = "c"
        elif self is DiffActions.DELETE:
            diff_type = "d"
        else:
            diff_type = "s"
        return diff_type



import pytest
from types import SimpleNamespace

from yamlpath.func import get_yaml_editor, get_yaml_data
from yamlpath.merger.enums import (
    AnchorConflictResolutions,
    AoHMergeOpts,
    ArrayMergeOpts,
    HashMergeOpts,
    MultiDocModes,
    OutputDocTypes,
    SetMergeOpts,
)
from yamlpath.wrappers import NodeCoords
from yamlpath.merger import MergerConfig
from yamlpath import YAMLPath
from tests.conftest import (
    info_warn_logger,
    quiet_logger,
    create_temp_yaml_file
)

class Test_merger_MergerConfig():
    """Tests for the MergerConfig class."""

    ###
    # get_insertion_point
    ###
    def test_get_insertion_point_default(self, quiet_logger):
        mc = MergerConfig(quiet_logger, SimpleNamespace())
        assert mc.get_insertion_point() == YAMLPath("/")

    def test_get_insertion_point_cli(self, quiet_logger):
        mc = MergerConfig(quiet_logger, SimpleNamespace(mergeat="la.tee.dah"))
        assert mc.get_insertion_point() == YAMLPath("/la/tee/dah")


    ###
    # get_document_format
    ###
    def test_get_document_format(self, quiet_logger):
        mc = MergerConfig(quiet_logger, SimpleNamespace())
        assert mc.get_document_format() == OutputDocTypes.AUTO


    ###
    # get_multidoc_mode
    ###
    def test_get_multidoc_mode_default(self, quiet_logger):
        mc = MergerConfig(quiet_logger, SimpleNamespace())
        assert mc.get_multidoc_mode() == MultiDocModes.CONDENSE_ALL

    def test_get_multidoc_mode_cli(self, quiet_logger):
        mc = MergerConfig(quiet_logger, SimpleNamespace(multi_doc_mode="matrix_merge"))
        assert mc.get_multidoc_mode() == MultiDocModes.MATRIX_MERGE


    ###
    # anchor_merge_mode
    ###
    def test_anchor_merge_mode_default(self, quiet_logger):
        mc = MergerConfig(quiet_logger, SimpleNamespace(anchors=None))
        assert mc.anchor_merge_mode() == AnchorConflictResolutions.STOP

    @pytest.mark.parametrize("setting, mode", [
        ("left", AnchorConflictResolutions.LEFT),
        ("rename", AnchorConflictResolutions.RENAME),
        ("right", AnchorConflictResolutions.RIGHT),
        ("stop", AnchorConflictResolutions.STOP),
    ])
    def test_anchor_merge_mode_cli(self, quiet_logger, setting, mode):
        mc = MergerConfig(quiet_logger, SimpleNamespace(anchors=setting))
        assert mc.anchor_merge_mode() == mode

    @pytest.mark.parametrize("setting, mode", [
        ("left", AnchorConflictResolutions.LEFT),
        ("rename", AnchorConflictResolutions.RENAME),
        ("right", AnchorConflictResolutions.RIGHT),
        ("stop", AnchorConflictResolutions.STOP),
    ])
    def test_anchor_merge_mode_ini(
        self, quiet_logger, tmp_path_factory, setting, mode
    ):
        config_file = create_temp_yaml_file(tmp_path_factory, """
        [defaults]
        anchors = {}
        """.format(setting))
        mc = MergerConfig(quiet_logger, SimpleNamespace(
            config=config_file
            , anchors=None))
        assert mc.anchor_merge_mode() == mode

    @pytest.mark.parametrize("cli, ini, mode", [
        ("left", "stop", AnchorConflictResolutions.LEFT),
        ("rename", "stop", AnchorConflictResolutions.RENAME),
        ("right", "stop", AnchorConflictResolutions.RIGHT),
        ("stop", "rename", AnchorConflictResolutions.STOP),
    ])
    def test_anchor_merge_mode_cli_overrides_ini(
        self, quiet_logger, tmp_path_factory, cli, ini, mode
    ):
        config_file = create_temp_yaml_file(tmp_path_factory, """
        [defaults]
        anchors = {}
        """.format(ini))
        mc = MergerConfig(quiet_logger, SimpleNamespace(
            config=config_file
            , anchors=cli))
        assert mc.anchor_merge_mode() == mode

    ###
    # hash_merge_mode
    ###
    def test_hash_merge_mode_default(self, quiet_logger):
        mc = MergerConfig(quiet_logger, SimpleNamespace(hashes=None))
        assert mc.hash_merge_mode(
            NodeCoords(None, None, None)) == HashMergeOpts.DEEP

    @pytest.mark.parametrize("setting, mode", [
        ("deep", HashMergeOpts.DEEP),
        ("left", HashMergeOpts.LEFT),
        ("right", HashMergeOpts.RIGHT),
    ])
    def test_hash_merge_mode_cli(self, quiet_logger, setting, mode):
        mc = MergerConfig(quiet_logger, SimpleNamespace(hashes=setting))
        assert mc.hash_merge_mode(
            NodeCoords(None, None, None)) == mode

    @pytest.mark.parametrize("setting, mode", [
        ("deep", HashMergeOpts.DEEP),
        ("left", HashMergeOpts.LEFT),
        ("right", HashMergeOpts.RIGHT),
    ])
    def test_hash_merge_mode_ini(
        self, quiet_logger, tmp_path_factory, setting, mode
    ):
        config_file = create_temp_yaml_file(tmp_path_factory, """
        [defaults]
        hashes = {}
        """.format(setting))
        mc = MergerConfig(quiet_logger, SimpleNamespace(
            config=config_file
            , hashes=None))
        assert mc.hash_merge_mode(
            NodeCoords(None, None, None)) == mode

    @pytest.mark.parametrize("cli, ini, mode", [
        ("deep", "left", HashMergeOpts.DEEP),
        ("left", "right", HashMergeOpts.LEFT),
        ("right", "deep", HashMergeOpts.RIGHT),
    ])
    def test_hash_merge_mode_cli_overrides_ini_defaults(
        self, quiet_logger, tmp_path_factory, cli, ini, mode
    ):
        config_file = create_temp_yaml_file(tmp_path_factory, """
        [defaults]
        hashes = {}
        """.format(ini))
        mc = MergerConfig(quiet_logger, SimpleNamespace(
            config=config_file
            , hashes=cli))
        assert mc.hash_merge_mode(
            NodeCoords(None, None, None)) == mode

    @pytest.mark.parametrize("cli, ini_default, ini_rule, mode", [
        ("deep", "left", "right", HashMergeOpts.RIGHT),
        ("left", "right", "deep", HashMergeOpts.DEEP),
        ("right", "deep", "left", HashMergeOpts.LEFT),
    ])
    def test_hash_merge_mode_ini_rule_overrides_cli(
        self, quiet_logger, tmp_path_factory, cli, ini_default, ini_rule, mode
    ):
        config_file = create_temp_yaml_file(tmp_path_factory, """
        [defaults]
        hashes = {}
        [rules]
        /hash = {}
        """.format(ini_default, ini_rule))
        lhs_yaml_file = create_temp_yaml_file(tmp_path_factory, """---
        hash:
          lhs_exclusive: lhs value 1
          merge_targets:
            subkey: lhs value 2
            subarray:
              - one
              - two
        array_of_hashes:
          - name: LHS Record 1
            id: 1
            prop: LHS value AoH 1
          - name: LHS Record 2
            id: 2
            prop: LHS value AoH 2
        """)
        lhs_yaml = get_yaml_editor()
        (lhs_data, lhs_loaded) = get_yaml_data(lhs_yaml, quiet_logger, lhs_yaml_file)

        mc = MergerConfig(quiet_logger, SimpleNamespace(
            config=config_file
            , hashes=cli))
        mc.prepare(lhs_data)

        node = lhs_data["hash"]
        parent = lhs_data
        parentref = "hash"

        assert mc.hash_merge_mode(
            NodeCoords(node, parent, parentref)) == mode


    ###
    # array_merge_mode
    ###
    def test_array_merge_mode_default(self, quiet_logger):
        mc = MergerConfig(quiet_logger, SimpleNamespace(arrays=None))
        assert mc.array_merge_mode(
            NodeCoords(None, None, None)) == ArrayMergeOpts.ALL

    @pytest.mark.parametrize("setting, mode", [
        ("all", ArrayMergeOpts.ALL),
        ("left", ArrayMergeOpts.LEFT),
        ("right", ArrayMergeOpts.RIGHT),
        ("unique", ArrayMergeOpts.UNIQUE),
    ])
    def test_array_merge_mode_cli(self, quiet_logger, setting, mode):
        mc = MergerConfig(quiet_logger, SimpleNamespace(arrays=setting))
        assert mc.array_merge_mode(
            NodeCoords(None, None, None)) == mode

    @pytest.mark.parametrize("setting, mode", [
        ("all", ArrayMergeOpts.ALL),
        ("left", ArrayMergeOpts.LEFT),
        ("right", ArrayMergeOpts.RIGHT),
        ("unique", ArrayMergeOpts.UNIQUE),
    ])
    def test_array_merge_mode_ini(
        self, quiet_logger, tmp_path_factory, setting, mode
    ):
        config_file = create_temp_yaml_file(tmp_path_factory, """
        [defaults]
        arrays = {}
        """.format(setting))
        mc = MergerConfig(quiet_logger, SimpleNamespace(
            config=config_file
            , arrays=None))
        assert mc.array_merge_mode(
            NodeCoords(None, None, None)) == mode

    @pytest.mark.parametrize("cli, ini, mode", [
        ("all", "left", ArrayMergeOpts.ALL),
        ("left", "right", ArrayMergeOpts.LEFT),
        ("right", "unique", ArrayMergeOpts.RIGHT),
        ("unique", "all", ArrayMergeOpts.UNIQUE),
    ])
    def test_array_merge_mode_cli_overrides_ini_defaults(
        self, quiet_logger, tmp_path_factory, cli, ini, mode
    ):
        config_file = create_temp_yaml_file(tmp_path_factory, """
        [defaults]
        arrays = {}
        """.format(ini))
        mc = MergerConfig(quiet_logger, SimpleNamespace(
            config=config_file
            , arrays=cli))
        assert mc.array_merge_mode(
            NodeCoords(None, None, None)) == mode

    @pytest.mark.parametrize("cli, ini_default, ini_rule, mode", [
        ("all", "left", "right", ArrayMergeOpts.RIGHT),
        ("left", "right", "unique", ArrayMergeOpts.UNIQUE),
        ("right", "unique", "all", ArrayMergeOpts.ALL),
        ("unique", "all", "left", ArrayMergeOpts.LEFT),
    ])
    def test_array_merge_mode_ini_rule_overrides_cli(
        self, quiet_logger, tmp_path_factory, cli, ini_default, ini_rule, mode
    ):
        config_file = create_temp_yaml_file(tmp_path_factory, """
        [defaults]
        arrays = {}
        [rules]
        /hash/merge_targets/subarray = {}
        """.format(ini_default, ini_rule))
        lhs_yaml_file = create_temp_yaml_file(tmp_path_factory, """---
        hash:
          lhs_exclusive: lhs value 1
          merge_targets:
            subkey: lhs value 2
            subarray:
              - one
              - two
        array_of_hashes:
          - name: LHS Record 1
            id: 1
            prop: LHS value AoH 1
          - name: LHS Record 2
            id: 2
            prop: LHS value AoH 2
        """)
        lhs_yaml = get_yaml_editor()
        (lhs_data, lhs_loaded) = get_yaml_data(lhs_yaml, quiet_logger, lhs_yaml_file)

        mc = MergerConfig(quiet_logger, SimpleNamespace(
            config=config_file
            , arrays=cli))
        mc.prepare(lhs_data)

        node = lhs_data["hash"]["merge_targets"]["subarray"]
        parent = lhs_data["hash"]["merge_targets"]
        parentref = "subarray"

        assert mc.array_merge_mode(
            NodeCoords(node, parent, parentref)) == mode


    ###
    # aoh_merge_mode
    ###
    def test_aoh_merge_mode_default(self, quiet_logger):
        mc = MergerConfig(quiet_logger, SimpleNamespace(aoh=None))
        assert mc.aoh_merge_mode(
            NodeCoords(None, None, None)) == AoHMergeOpts.ALL

    @pytest.mark.parametrize("setting, mode", [
        ("all", AoHMergeOpts.ALL),
        ("deep", AoHMergeOpts.DEEP),
        ("left", AoHMergeOpts.LEFT),
        ("right", AoHMergeOpts.RIGHT),
        ("unique", AoHMergeOpts.UNIQUE),
    ])
    def test_aoh_merge_mode_cli(self, quiet_logger, setting, mode):
        mc = MergerConfig(quiet_logger, SimpleNamespace(aoh=setting))
        assert mc.aoh_merge_mode(
            NodeCoords(None, None, None)) == mode

    @pytest.mark.parametrize("setting, mode", [
        ("all", AoHMergeOpts.ALL),
        ("deep", AoHMergeOpts.DEEP),
        ("left", AoHMergeOpts.LEFT),
        ("right", AoHMergeOpts.RIGHT),
        ("unique", AoHMergeOpts.UNIQUE),
    ])
    def test_aoh_merge_mode_ini(
        self, quiet_logger, tmp_path_factory, setting, mode
    ):
        config_file = create_temp_yaml_file(tmp_path_factory, """
        [defaults]
        aoh = {}
        """.format(setting))
        mc = MergerConfig(quiet_logger, SimpleNamespace(
            config=config_file
            , aoh=None))
        assert mc.aoh_merge_mode(
            NodeCoords(None, None, None)) == mode

    @pytest.mark.parametrize("cli, ini, mode", [
        ("all", "deep", AoHMergeOpts.ALL),
        ("deep", "left", AoHMergeOpts.DEEP),
        ("left", "right", AoHMergeOpts.LEFT),
        ("right", "unique", AoHMergeOpts.RIGHT),
        ("unique", "all", AoHMergeOpts.UNIQUE),
    ])
    def test_aoh_merge_mode_cli_overrides_ini_defaults(
        self, quiet_logger, tmp_path_factory, cli, ini, mode
    ):
        config_file = create_temp_yaml_file(tmp_path_factory, """
        [defaults]
        aoh = {}
        """.format(ini))
        mc = MergerConfig(quiet_logger, SimpleNamespace(
            config=config_file
            , aoh=cli))
        assert mc.aoh_merge_mode(
            NodeCoords(None, None, None)) == mode

    @pytest.mark.parametrize("cli, ini_default, ini_rule, mode", [
        ("all", "deep", "left", AoHMergeOpts.LEFT),
        ("deep", "left", "right", AoHMergeOpts.RIGHT),
        ("left", "right", "unique", AoHMergeOpts.UNIQUE),
        ("right", "unique", "all", AoHMergeOpts.ALL),
        ("unique", "all", "deep", AoHMergeOpts.DEEP),
    ])
    def test_aoh_merge_mode_ini_rule_overrides_cli(
        self, quiet_logger, tmp_path_factory, cli, ini_default, ini_rule, mode
    ):
        config_file = create_temp_yaml_file(tmp_path_factory, """
        [defaults]
        aoh = {}
        [rules]
        /array_of_hashes = {}
        """.format(ini_default, ini_rule))
        lhs_yaml_file = create_temp_yaml_file(tmp_path_factory, """---
        hash:
          lhs_exclusive: lhs value 1
          merge_targets:
            subkey: lhs value 2
            subarray:
              - one
              - two
        array_of_hashes:
          - name: LHS Record 1
            id: 1
            prop: LHS value AoH 1
          - name: LHS Record 2
            id: 2
            prop: LHS value AoH 2
        """)
        lhs_yaml = get_yaml_editor()
        (lhs_data, lhs_loaded) = get_yaml_data(lhs_yaml, quiet_logger, lhs_yaml_file)

        mc = MergerConfig(quiet_logger, SimpleNamespace(
            config=config_file
            , aoh=cli))
        mc.prepare(lhs_data)

        node = lhs_data["array_of_hashes"]
        parent = lhs_data
        parentref = "array_of_hashes"

        assert mc.aoh_merge_mode(
            NodeCoords(node, parent, parentref)) == mode


    ###
    # aoh_merge_key
    ###
    def test_aoh_merge_key_default(self, quiet_logger, tmp_path_factory):
        lhs_yaml_file = create_temp_yaml_file(tmp_path_factory, """---
        hash:
          lhs_exclusive: lhs value 1
          merge_targets:
            subkey: lhs value 2
            subarray:
              - one
              - two
        array_of_hashes:
          - name: LHS Record 1
            id: 1
            prop: LHS value AoH 1
          - name: LHS Record 2
            id: 2
            prop: LHS value AoH 2
        """)
        lhs_yaml = get_yaml_editor()
        (lhs_data, lhs_loaded) = get_yaml_data(lhs_yaml, quiet_logger, lhs_yaml_file)

        mc = MergerConfig(quiet_logger, SimpleNamespace())
        mc.prepare(lhs_data)

        node = lhs_data["array_of_hashes"]
        parent = lhs_data
        parentref = "array_of_hashes"
        record = node[0]

        assert mc.aoh_merge_key(
            NodeCoords(node, parent, parentref), record) == "name"

    def test_aoh_merge_key_ini(self, quiet_logger, tmp_path_factory):
        config_file = create_temp_yaml_file(tmp_path_factory, """
        [keys]
        /array_of_hashes = id
        """)
        lhs_yaml_file = create_temp_yaml_file(tmp_path_factory, """---
        hash:
          lhs_exclusive: lhs value 1
          merge_targets:
            subkey: lhs value 2
            subarray:
              - one
              - two
        array_of_hashes:
          - name: LHS Record 1
            id: 1
            prop: LHS value AoH 1
          - name: LHS Record 2
            id: 2
            prop: LHS value AoH 2
        """)
        lhs_yaml = get_yaml_editor()
        (lhs_data, lhs_loaded) = get_yaml_data(lhs_yaml, quiet_logger, lhs_yaml_file)

        mc = MergerConfig(quiet_logger, SimpleNamespace(config=config_file))
        mc.prepare(lhs_data)

        node = lhs_data["array_of_hashes"]
        parent = lhs_data
        parentref = "array_of_hashes"
        record = node[0]

        assert mc.aoh_merge_key(
            NodeCoords(node, parent, parentref), record) == "id"

    def test_aoh_merge_key_ini_inferred_parent(
        self, quiet_logger, tmp_path_factory
    ):
        config_file = create_temp_yaml_file(tmp_path_factory, """
        [keys]
        /array_of_hashes = prop
        """)
        lhs_yaml_file = create_temp_yaml_file(tmp_path_factory, """---
        hash:
          lhs_exclusive: lhs value 1
          merge_targets:
            subkey: lhs value 2
            subarray:
              - one
              - two
        array_of_hashes:
          - name: LHS Record 1
            id: 1
            prop: LHS value AoH 1
          - name: LHS Record 2
            id: 2
            prop: LHS value AoH 2
        """)
        lhs_yaml = get_yaml_editor()
        (lhs_data, lhs_loaded) = get_yaml_data(lhs_yaml, quiet_logger, lhs_yaml_file)

        mc = MergerConfig(quiet_logger, SimpleNamespace(config=config_file))
        mc.prepare(lhs_data)

        node = lhs_data["array_of_hashes"][1]
        parent = lhs_data["array_of_hashes"]
        parentref = 1
        record = node

        assert mc.aoh_merge_key(
            NodeCoords(node, parent, parentref), record) == "prop"


    ###
    # set_merge_mode
    ###
    def test_set_merge_mode_default(self, quiet_logger):
        mc = MergerConfig(quiet_logger, SimpleNamespace(sets=None))
        assert mc.set_merge_mode(
            NodeCoords(None, None, None)) == SetMergeOpts.UNIQUE

    @pytest.mark.parametrize("setting, mode", [
        ("left", SetMergeOpts.LEFT),
        ("right", SetMergeOpts.RIGHT),
        ("unique", SetMergeOpts.UNIQUE),
    ])
    def test_set_merge_mode_cli(self, quiet_logger, setting, mode):
        mc = MergerConfig(quiet_logger, SimpleNamespace(sets=setting))
        assert mc.set_merge_mode(
            NodeCoords(None, None, None)) == mode

    @pytest.mark.parametrize("setting, mode", [
        ("left", SetMergeOpts.LEFT),
        ("right", SetMergeOpts.RIGHT),
        ("unique", SetMergeOpts.UNIQUE),
    ])
    def test_set_merge_mode_ini(
        self, quiet_logger, tmp_path_factory, setting, mode
    ):
        config_file = create_temp_yaml_file(tmp_path_factory, """
        [defaults]
        sets = {}
        """.format(setting))
        mc = MergerConfig(quiet_logger, SimpleNamespace(
            config=config_file
            , sets=None))
        assert mc.set_merge_mode(
            NodeCoords(None, None, None)) == mode

    @pytest.mark.parametrize("cli, ini, mode", [
        ("left", "right", SetMergeOpts.LEFT),
        ("right", "unique", SetMergeOpts.RIGHT),
        ("unique", "all", SetMergeOpts.UNIQUE),
    ])
    def test_set_merge_mode_cli_overrides_ini_defaults(
        self, quiet_logger, tmp_path_factory, cli, ini, mode
    ):
        config_file = create_temp_yaml_file(tmp_path_factory, """
        [defaults]
        sets = {}
        """.format(ini))
        mc = MergerConfig(quiet_logger, SimpleNamespace(
            config=config_file
            , sets=cli))
        assert mc.set_merge_mode(
            NodeCoords(None, None, None)) == mode

    @pytest.mark.parametrize("cli, ini_default, ini_rule, mode", [
        ("left", "right", "unique", SetMergeOpts.UNIQUE),
        ("right", "unique", "left", SetMergeOpts.LEFT),
        ("unique", "left", "right", SetMergeOpts.RIGHT),
    ])
    def test_set_merge_mode_ini_rule_overrides_cli(
        self, quiet_logger, tmp_path_factory, cli, ini_default, ini_rule, mode
    ):
        config_file = create_temp_yaml_file(tmp_path_factory, """
        [defaults]
        sets = {}
        [rules]
        /hash/merge_targets/subset = {}
        """.format(ini_default, ini_rule))
        lhs_yaml_file = create_temp_yaml_file(tmp_path_factory, """---
        hash:
          lhs_exclusive: lhs value 1
          merge_targets:
            subkey: lhs value 2
            subset:
              ? one
              ? two
        """)
        lhs_yaml = get_yaml_editor()
        (lhs_data, lhs_loaded) = get_yaml_data(lhs_yaml, quiet_logger, lhs_yaml_file)

        mc = MergerConfig(quiet_logger, SimpleNamespace(
            config=config_file
            , sets=cli))
        mc.prepare(lhs_data)

        node = lhs_data["hash"]["merge_targets"]["subset"]
        parent = lhs_data["hash"]["merge_targets"]
        parentref = "subset"

        assert mc.set_merge_mode(
            NodeCoords(node, parent, parentref)) == mode


    ###
    # Edge Cases
    ###
    def test_warn_when_rules_matches_zero_nodes(
        self, capsys, info_warn_logger, tmp_path_factory
    ):
        config_file = create_temp_yaml_file(tmp_path_factory, """
        [rules]
        /does_not_exist = left
        /array_of_hashes[name = "Does Not Compute"] = right
        """)
        lhs_yaml_file = create_temp_yaml_file(tmp_path_factory, """---
        hash:
          lhs_exclusive: lhs value 1
          merge_targets:
            subkey: lhs value 2
            subarray:
              - one
              - two
        array_of_hashes:
          - name: LHS Record 1
            id: 1
            prop: LHS value AoH 1
          - name: LHS Record 2
            id: 2
            prop: LHS value AoH 2
        """)
        lhs_yaml = get_yaml_editor()
        lhs_data = get_yaml_data(lhs_yaml, info_warn_logger, lhs_yaml_file)

        mc = MergerConfig(info_warn_logger, SimpleNamespace(config=config_file))
        mc.prepare(lhs_data)

        console = capsys.readouterr()
        assert "YAML Path matches no nodes" in console.out



"""
Implement DiffEntry.

Copyright 2020 William W. Kimball, Jr. MBA MSIS
"""
import json
from typing import Any

from ruamel.yaml.comments import CommentedBase, TaggedScalar

from yamlpath.common import Parsers
from yamlpath.enums import PathSeperators
from yamlpath import YAMLPath
from .enums.diffactions import DiffActions


class DiffEntry:
    """One entry of a diff."""

    def __init__(
        self, action: DiffActions, path: YAMLPath, lhs: Any, rhs: Any,
        **kwargs
    ):
        """
        Instantiate a new DiffEntry.

        Parameters:
        1. action (DiffAction) The action taken for one document to become the
            next
        2. path (YAMLPath) Location within the LHS document which changes to
            becomes the RHS document
        3. lhs (Any) The Left-Hand-Side (original) document
        4. rhs (Any) The Right-Hand-Side (altered) document

        Keyword Arguments:
        * lhs_iteration (Any) "Rough" position of the original element within
          its document before it was changed
        * lhs_parent (Any) Parent of the original data element
        * rhs_iteration (Any) "Rough" position of the changed element within
          its document, if it existed before the change (otherwise it'll be 0s)
        * rhs_parent (Any) Parent of the changed data element

        Returns:  N/A
        """
        self._action: DiffActions = action
        self._path: YAMLPath = path
        self._lhs: Any = lhs
        self._rhs: Any = rhs
        self._key_tag = kwargs.pop("key_tag", None)
        self._set_index(lhs, rhs, **kwargs)
        self._verbose = False

    def _set_index(self, lhs: Any, rhs: Any, **kwargs) -> None:
        """
        Build the sortable index for this entry.

        Parameters:
        1. lhs (Any) The Left-Hand-Side (original) document
        2. rhs (Any) The Right-Hand-Side (altered) document

        Keyword Arguments:
        * lhs_iteration (Any) "Rough" position of the original element within
          its document before it was changed
        * lhs_parent (Any) Parent of the original data element
        * rhs_iteration (Any) "Rough" position of the changed element within
          its document, if it existed before the change (otherwise it'll be 0s)
        * rhs_parent (Any) Parent of the changed data element

        Returns:  N/A
        """
        lhs_lc = DiffEntry._get_index(lhs, kwargs.pop("lhs_parent", None))
        rhs_lc = DiffEntry._get_index(rhs, kwargs.pop("rhs_parent", None))
        lhs_iteration = kwargs.pop("lhs_iteration", 0)
        rhs_iteration = kwargs.pop("rhs_iteration", 0)
        lhs_iteration = 0 if lhs_iteration is None else lhs_iteration
        rhs_iteration = 0 if rhs_iteration is None else rhs_iteration
        lhs_line = float(lhs_lc)
        if lhs_line == 0.0 or self.action is DiffActions.ADD:
            lhs_lc, rhs_lc = rhs_lc, lhs_lc
        self._index = "{}.{}.{}.{}".format(
            lhs_lc, lhs_iteration, rhs_lc, rhs_iteration)

    def __str__(self) -> str:
        """Get the string representation of this object."""
        diffaction = self._action
        path = self._path if self._path else "-"
        key_tag = ""
        if self._key_tag:
            key_tag = " {}".format(self._key_tag)
        output = "{}{} {}{}\n".format(
            diffaction, self._index if self.verbose else "", path, key_tag)
        if diffaction is DiffActions.ADD:
            output += DiffEntry._present_data(self._rhs, ">")
        elif diffaction is DiffActions.CHANGE:
            output += "{}\n---\n{}".format(
                DiffEntry._present_data(self._lhs, "<"),
                DiffEntry._present_data(self._rhs, ">"))
        elif diffaction is DiffActions.DELETE:
            output += "{}".format(DiffEntry._present_data(self._lhs, "<"))
        else:
            output += "{}".format(DiffEntry._present_data(self._lhs, "="))
        return output

    @property
    def action(self) -> DiffActions:
        """Get the action of this difference (read-only)."""
        return self._action

    @property
    def path(self) -> YAMLPath:
        """Get the YAML Path of this difference (read-only)."""
        return self._path

    @property
    def lhs(self) -> Any:
        """Get the LHS value of this difference (read-only)."""
        return self._lhs

    @property
    def index(self) -> str:
        """Get the sortable index for this entry (read-only)."""
        return self._index

    @property
    def pathsep(self) -> PathSeperators:
        """Seperator used to delimit reported YAML Paths (accessor)."""
        return self._path.seperator

    @pathsep.setter
    def pathsep(self, value: PathSeperators) -> None:
        """Seperator used to delimit reported YAML Paths (mutator)."""
        # No unnecessary changes
        if value is not self.pathsep:
            self._path.seperator = value

    @property
    def verbose(self) -> bool:
        """Output verbosity (accessor)."""
        return self._verbose

    @verbose.setter
    def verbose(self, value: bool) -> None:
        """Output verbosity (mutator)."""
        # No unnecessary changes
        if value != self.verbose:
            self._verbose = value

    @classmethod
    def _get_lc(cls, data: Any) -> str:
        """Get the line.column of a data element."""
        data_lc = "0.0"
        if isinstance(data, CommentedBase):
            dlc = data.lc
            data_lc = "{}.{}".format(
                dlc.line if dlc.line is not None else 0,
                dlc.col if dlc.col is not None else 0
            )
        return data_lc

    @classmethod
    def _get_index(cls, data: Any, parent: Any) -> str:
        """Get the document index of a data element."""
        data_lc = DiffEntry._get_lc(data)
        if data_lc == "0.0":
            data_lc = DiffEntry._get_lc(parent)
        return data_lc

    @classmethod
    def _present_data(cls, data: Any, prefix: str) -> str:
        """Stringify data."""
        json_safe_data = Parsers.jsonify_yaml_data(data)
        formatted_data = json_safe_data
        if isinstance(json_safe_data, str):
            formatted_data = json_safe_data.strip()
        json_data = json.dumps(
            formatted_data).replace("\\n", "\n{} ".format(prefix))
        data_tag = ""
        if isinstance(data, TaggedScalar) and data.tag.value:
            data_tag = "{} ".format(data.tag.value)
        return "{} {}{}".format(prefix, data_tag, json_data)
